#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_set_cover.py.

Contains the SetCover class. See ``help(qubovert.problems.SetCover)``.

"""

from numpy import log2, allclose
from qubovert.utils import (
    QUBOMatrix, solve_qubo_bruteforce, solution_type, spin_to_boolean
)
from qubovert.problems import Problem


__all__ = 'SetCover',


class SetCover(Problem):
    """SetCover.

    Class to manage converting (Weighted) Set Cover to and from its QUBO and
    QUSO formluations. Based on the paper "Ising formulations of many
    NP problems", hereforth designated as [Lucas].

    The goal of the SetCover problem is to find the smallest number of subsets
    of U in V such that union over the elements equals U. The goal of the
    Weghted SetCover problem is to find the smallest weight of subsets of U
    in V such that union over the elements equals U, where each element in V
    has an associated weight.

    This class inherits some methods and attributes from the
    ``qubovert.problems.Problem`` class.

    Example
    -------
    >>> from qubovert.problems import SetCover
    >>> from any_module import qubo_solver
    >>> # or you can use my bruteforce solver...
    >>> # from qubovert.utils import solve_qubo_bruteforce as qubo_solver
    >>> U = {"a", "b", "c", "d"}
    >>> V = [{"a", "b"}, {"a", "c"}, {"c", "d"}]
    >>> problem = SetCover(U, V)
    >>> Q = problem.to_qubo()
    >>> obj, sol = qubo_solver(Q)
    >>> solution = problem.convert_solution(sol)

    >>> print(solution)
    {0, 2}
    >>> print(problem.is_solution_valid(solution))
    True  # since V[0] + V[2] covers all of U
    >>> print(obj == len(solution))
    True

    """

    def __init__(self, U, V, weights=None, log_trick=True, M=None):
        """__init__.

        The goal of the SetCover problem is to find the smallest number of
        elements in V such that the union over the elements equals U. Weighted
        Set Cover is the task of finding the smallest weight of elements in V
        such that the union over the elements equals U, where each element in V
        has an associated weight. All naming conventions follow the names in
        the paper [Lucas].

        Parameters
        ----------
        U : set.
            The set of all elements to cover.
        V : iterable of sets.
            Each set is one of the subsets.
        weights : iterable of numbers (optional, defaults to None).
            Important: weights must be normalized such that
            `max(weights) == 1`; similarly, `len(weights) == len(V)`. These are
            the weights for the weighted Set Cover problem. If `weights` is
            left to `None`, then the problem will default to the unweighted Set
            Cover problem (ie all the weights equal to 1).
        log_trick : boolean (optional, defaults to True).
            Indicates whether or not to use the log trick discussed in [Lucas].
        M : int (optional, defaults to None). We recommend not adjusting this.
            The number of duplicated elements to allow for, see [Lucas]. If M
            is None, then it will be set to the value required to ensure
            accuracy of the model. To possibly sacrifice accuracy but decrease
            the number of variables in the model, you can set M to something
            smaller.

        """
        self._log_trick = log_trick
        self._U = U.copy()
        self._V = type(V)(x.copy() for x in V)
        self._N, self._n = len(self.V), len(self.U)

        if weights is None:
            self._weights = type(V)(1 for _ in V)
        else:
            if len(weights) != len(V):
                raise ValueError("`weights` must be the same length as `V`")
            elif not allclose(max(weights), 1):
                raise ValueError("`weights` must be normalized such that "
                                 "`max(weights) == 1`")
            self._weights = type(weights)(weights)  # copy weights

        if M is not None:
            self._M = M
        else:
            self._M = max(
                sum(int(alpha in v) for v in self.V)
                for alpha in self.U
            )
        self._log_M = int(log2(self._M))+1

        # map each alpha in U to a unique integer index. used for the QUBO and
        # quso conversions.
        self._alpha_to_index = {alpha: i for i, alpha in enumerate(self._U)}

    @property
    def U(self):
        """U.

        A copy of the U set. Updating the copy will not update the
        instance U.

        Return
        ------
        U : set.
            A copy of the set of all elements.

        """
        return self._U.copy()

    @property
    def V(self):
        """V.

        A copy of the V list/tuple. Updating the copy will not update the
        instance V.

        Return
        ------
        V : iterable of sets.
            A copy of the subsets.

        """
        return type(self._V)(x.copy() for x in self._V)

    @property
    def weights(self):
        """weights.

        A copy of the weights list/tuple. Updating the copy will not update the
        instance weights.

        Return
        ------
        weights : iterable of numbers, where ``max(weights) == 1``.

        """
        return type(self._weights)(self._weights)

    @property
    def M(self):
        """M.

        A copy of the value for the M variable. See [Lucas].

        Returns
        -------
        M : int.

        """
        return self._M

    @property
    def log_trick(self):
        """log_trick.

        A copy of the value for log_trick, indicating whether or not to use
        the log trick. See [Lucas].

        Return
        ------
        log_trick : bool.

        """
        return self._log_trick

    @property
    def num_binary_variables(self):
        """num_binary_variables.

        The number of binary variables that the QUBO and QUSO use.

        Return
        ------
        num :  int.
            The number of variables in the QUBO/QUSO formulation.

        """
        if self._log_trick:
            return self._N + self._n*(self._log_M+1)
        return self._N + self._n*self._M

    def is_coverable(self):
        """is_coverable.

        Returns whether or not there exists a valid solution. Ie if there
        exists a combination of subsets in V that cover U.

        Return
        ------
        coverable : bool.
            True if it is possible to construct a valid solution from V and U,
            False otherwise.

        Examples
        --------
        >>> U = {0, 1, 2}
        >>> V = [{0}, {0, 1}]
        >>> SetCover(U, V).is_coverable()
        False

        >>> U = {0, 1, 2}
        >>> V = [{0, 2}, {0, 1}]
        >>> SetCover(U, V).is_coverable()
        True

        """
        # assuming every subset is chosen, does this cover U?
        return self.is_solution_valid([1]*self._N)

    def _filtered_range(self, alpha, start=0):
        """_filtered_range.

        Find each set in self.V[start:] that contains alpha.

        Parameters
        ----------
        alpha : element in self.U.
        start : int (optional, defaults to 0)
            Must be less than ``len(self.V)``. Indicates at which subset to
            start looking at.

        Return
        ------
        f : filter object.
            The indices of V corresponding to subsets that contain alpha.

        """
        return filter(lambda k: alpha in self._V[k], range(start, self._N))

    def _x(self, alpha, m):
        r"""_x.

        Return a unique index for each of the ancilla variables.

        Parameters
        ----------
        alpha : element in self.U.
        m : int.
            Index of the ancilla binary variables :math:`x_{\alpha, m}` (see
            the Set Cover section in [Lucas]).

        Return
        ------
        i : int.
            Unique index of the ancilla variable.

        """
        mm = m if self._log_trick else m-1
        return self._N + self._alpha_to_index[alpha] + self._n*mm

    def to_qubo(self, A=2, B=1):
        r"""to_qubo.

        Create and return the set cover problem in QUBO form following section
        5.1 of [Lucas]. The Q matrix for the QUBO
        will be returned as an uppertriangular dictionary. Thus, the problem
        becomes minimizing :math:`\sum_{i \leq j} x_i x_j Q_{ij}`. A and B are
        parameters to enforce constraints.

        If all the constraints are satisfied, then the objective function
        will be equal to the total number of sets in the cover (or for
        the weighted Set Cover problem, it will equal the total weight
        of included sets in the cover).

        Parameters
        ----------
        A: positive float (optional, defaults to 2).
            A enforces the constraints. See section 5.1 of [Lucas].
        B: positive float that is less than A (optional, defaults to 1).
            See section 5.1 of [Lucas].

        Return
        ------
        Q : qubovert.utils.QUBOMatrix object.
            The upper triangular QUBO matrix, a QUBOMatrix object.
            For most practical purposes, you can use QUBOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.QUBOMatrix)``.

        """
        # all naming conventions follow the paper listed in the docstring

        Q = QUBOMatrix()

        Q += self._n * A  # constant comes from the constraints

        # encode H_B (equation 46)
        for i in range(self._N):
            Q[(i,)] += self._weights[i] * B

        # encode H_A

        for alpha in self._U:

            if not self._log_trick:  # (Equation 45)

                # first constraint
                for m in range(1, self._M+1):
                    i = self._x(alpha, m)
                    Q[(i, i)] -= A
                    for mp in range(m+1, self._M+1):
                        ip = self._x(alpha, mp)
                        Q[(i, ip)] += 2*A

                # second constraint
                for m in range(1, self._M+1):
                    i = self._x(alpha, m)
                    Q[(i, i)] += A*m*m
                    for mp in range(m+1, self._M+1):
                        ip = self._x(alpha, mp)
                        Q[(i, ip)] += 2*A*m*mp

                    for j in self._filtered_range(alpha):
                        Q[(j, i)] -= 2*A*m

            else:  # using the log_trick

                # no first constraint now, but modify the second constraint.
                for m in range(self._log_M+1):
                    i = self._x(alpha, m)
                    Q[(i, i)] += A * (pow(2, 2*m) + 2 * pow(2, m))
                    for mp in range(m+1, self._log_M+1):
                        ip = self._x(alpha, mp)
                        Q[(i, ip)] += 2*A*pow(2, m+mp)
                    for j in self._filtered_range(alpha):
                        Q[(j, i)] -= 2*A*pow(2, m)

            for i in self._filtered_range(alpha):
                Q[(i,)] += A if not self._log_trick else -A
                for j in self._filtered_range(alpha, i+1):
                    Q[(i, j)] += 2 * A

        return Q

    def convert_solution(self, solution, spin=False):
        """convert_solution.

        Convert the solution to the QUBO or QUSO to the solution to the Set
        Cover problem.

        Parameters
        ----------
        solution : iterable or dict.
            The QUBO or QUSO solution output. The QUBO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or -1 or 1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool (optional, defaults to False).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Returns
        -------
        res : set.
            A set of which sets are included in the set cover. So if this
            function returns ``{0, 2, 3}``, then the set cover is the sets
            ``V[0]``, ``V[2]``, and ``V[3]``.

        """
        sol_type = solution_type(solution, 'spin' if spin else 'bool')
        if sol_type == 'spin':
            solution = spin_to_boolean(solution)
        return set(i for i in range(self._N) if solution[i])

    def is_solution_valid(self, solution, spin=False):
        """is_solution_valid.

        Returns whether or not the proposed solution covers all the elements in
        U.

        Parameters
        ----------
        solution : iterable or dict.
            solution can be the output of SetCover.convert_solution,
            or the  QUBO or QUSO solver output. The QUBO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool (optional, defaults to False).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        valid : boolean.
            True if the proposed solution is valid, else False.

        """
        if not isinstance(solution, set):
            solution = self.convert_solution(solution, spin)

        covered = set(x for i in solution for x in self._V[i])
        return covered == self._U

    def solve_bruteforce(self, all_solutions=False):
        """solve_bruteforce.

        Solves the Set Cover problem exactly with a brute force method. THIS
        SHOULD NOT BE USED FOR LARGE PROBLEMS! The advantage over this method
        as opposed to using a brute force QUBO solver is that the QUBO
        formulation has many slack variables.

        Parameters
        ----------
        all_solutions : boolean (optional, defaults to False).
            If ``all_solutions`` is set to True, all the best solutions to the
            problem will be returned rather than just one of the best. If the
            problem is very big, then it is best if ``all_solutions == False``,
            otherwise this function will use a lot of memory.

        Returns
        -------
        res : set or list of sets.
            A set of which sets are included in the set cover. So if this
            function returns ``{0, 2, 3}``, then the set cover is the sets
            ``V[0]``, ``V[2]``, and ``V[3]``. If ``all_solutions == True``,
            then ``res`` will be a list of sets, where each element of the list
            is one of the optimal solutions.

        Examples
        --------
        >>> from qubovert import SetCover
        >>> U = {"a", "b", "c", "d"}
        >>> V = [{"a", "b"}, {"a", "c"}, {"c", "d"}]
        >>> problem = SetCover(U, V)
        >>> print(problem.solve_bruteforce())
        {0, 2}

        """
        # minimize the objective function
        Q = {(i,): self._weights[i] for i in range(self._N)}
        # subject to the constraint
        valid = self.is_solution_valid

        obj, sol = solve_qubo_bruteforce(Q, all_solutions, valid)

        if obj is None:
            raise ValueError("Problem is not solvable. See "
                             "``SetCover.is_coverable()``")

        if all_solutions:
            return [self.convert_solution(s) for s in sol]
        return self.convert_solution(sol)

#        best = None
#        all_sols = {}
#        for x in range(1 << self._N):
#            sol = decimal_to_boolean(x, self._N)
#            cover = self.convert_solution(sol)
#            if self.is_solution_valid(cover):
#                if not all_solutions and (best is None or
#                                          len(cover) < len(best)):
#                    best = cover
#                elif all_solutions and (best is None or
#                                        len(cover) <= len(best)):
#                    best = cover
#                    all_sols.setdefault(len(cover), []).append(cover)
#
#        if best is None:
#            raise ValueError("Problem is not solvable. See "
#                             "``SetCover.is_coverable()``")
#        if all_solutions:
#            return all_sols[len(best)]
#        return best



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_normalize.py.

This file contains the function to normalize QUBOs, PUBOs, QUSOs,
PUSOs, etc.

"""


__all__ = 'normalize',


def normalize(D, value=1):
    """normalize.

    Normalize the coefficients to a maximum magnitude.

    Parameters
    ----------
    D : dict or subclass of dict.
    value : float (optional, defaults to 1).
        Every coefficient value will be normalized such that the
        coefficient with the maximum magnitude will be +/- 1.

    Return
    ------
    res : same as type(D).
        ``D`` but with coefficients that are normalized to be within +/- value.

    Examples
    --------
    >>> from qubovert.utils import DictArithmetic, normalize
    >>> d = {(0, 1): 1, (1, 2, 'x'): 4}
    >>> print(normalize(d))
    {(0, 1): 0.25, (1, 2, 'x'): 1}

    >>> from qubovert.utils import DictArithmetic, normalize
    >>> d = {(0, 1): 1, (1, 2, 'x'): -4}
    >>> print(normalize(d))
    {(0, 1): 0.25, (1, 2, 'x'): -1}

    >>> from qubovert import PUBO
    >>> d = PUBO({(0, 1): 1, (1, 2, 'x'): 4})
    >>> print(normalize(d))
    {(0, 1): 0.25, (1, 2, 'x'): 1}

    >>> from qubovert.utils import PUBO
    >>> d = PUBO({(0, 1): 1, (1, 2, 'x'): -4})
    >>> print(normalize(d))
    {(0, 1): 0.25, (1, 2, 'x'): -1}


    """
    res = type(D)()
    mult = value / max(abs(v) for v in D.values())
    for k, v in D.items():
        res[k] = mult * v
    return res



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Defines the version number and details of qubovert."""

__all__ = (
    '__version__', '__author__', '__authoremail__', '__license__',
    '__sourceurl__', '__docsurl__', '__description__'
)


__version__ = "1.1.5"
__author__ = "Joseph T. Iosue"
__authoremail__ = "joe.iosue@yahoo.com"
__license__ = "Apache Software License 2.0"
__sourceurl__ = "https://github.com/jiosue/qubovert"
__docsurl__ = "https://qubovert.readthedocs.io"
__description__ = (
    "A package for formulating, simulating, and solving "
    "problems in boolean and spin form"
)



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_puso.py.

Contains the PUSO class. See ``help(qubovert.PUSO)``.

"""

from .utils import BO, PUSOMatrix, puso_to_pubo
from . import QUSO


__all__ = 'PUSO',


class PUSO(BO, PUSOMatrix):
    """PUSO.

    Class to manage converting general PUSO problems to and from their
    PUSO, PUBO, QUSO, and QUBO formluations. In general, this class
    deals with unconstrained optimization problems that have arbitrary degree.
    To convert this to a QUSO (see ``to_quso``) or QUBO (``to_qubo``) we have
    to introduce ancilla variables. The ``convert_solution`` method deals with
    converting a solution to the problem with ancilla variables back to the
    solution to the original problem.

    This class deals with PUSOs that have spin labels that do not range
    from 0 to n-1. Note that it is generally
    more efficient to initialize an empty PUSO object and then build the
    PUSO, rather than initialize a PUSO object with an already built dict.

    PUSO inherits some methods and attributes the ``PUSOMatrix`` class. See
    ``help(qubovert.utils.PUSOMatrix)``.

    PUSO inherits some methods and attributes the ``BO`` class. See
    ``help(qubovert.utils.BO)``.

    Example
    -------
    >>> puso = PUSO()
    >>> puso[('a',)] += 5
    >>> puso[(0, 'a', 1)] -= 2
    >>> puso -= 1.5
    >>> puso
    {('a',): 5, ('a', 0, 1): -2, (): -1.5}

    >>> puso = PUSO({('a',): 5, (0, 'a', 1): -2, (): -1.5})
    >>> puso
    {('a',): 5, ('a', 0, 1): -2, (): -1.5}
    >>> H = puso.to_puso()
    >>> H
    {(0,): 5, (0, 1): -2, (): -1.5}
    >>> puso.convert_solution({0: 1, 1: -1, 2: 1})
    {'a': 1, 0: -1, 1: 1}

    In the next example, notice that we introduce ancilla variables to
    represent that ```(0, 1)`` term. See the ``to_quso`` method for more info.

    >>> puso = PUSO({('a',): 5, (0, 'a', 1): -2, (): -1.5})
    >>> puso.mapping
    {'a': 0, 0: 1, 1: 2}
    >>> L = puso.to_quso()
    >>> L
    {(0,): 0.75, (): 11.25, (3,): 8.5, (0, 1): 4.25, (1,): -4.25, (0, 3): -8.5,
     (1, 3): -8.5, (2, 3): -4.0, (1, 2): 2.0, (0, 2): 2.0, (2,): -2.0}
    >>> puso.convert_solution({0: 1, 1: -1, 2: 1, 3: -1})
    {'a': 1, 0: -1, 1: 1}

    Note
    ----
    Note that keys will end up sorted by their hash. Hashes will not be
    consistent across Python sessions (unless they are integers)! For example,
    both of the following can happen:

    >>> print(PUSO({('a', 0): 1, (0, 1): -1}))
    {('a', 0): 1, (0, 1): -1}

    or

    >>> print(PUSO({('a', 0): 1, (0, 1): -1}))
    {(0, 'a'): 1, (0, 1): -1}

    But the following will never happen:

    >>> print(PUSO({('a', 0): 1, (0, 1): -1}))
    {('a', 0): 1, (1, 0): -1}

    Ie integers will always be correctly sorted.

    Note
    ----
    For efficiency, many internal variables including mappings are computed as
    the problemis being built. This can cause these
    values to be wrong for some specific situations. Calling ``refresh``
    will rebuild the dictionary, resetting all of the values. See
    ``help(PUSO.refresh)``

    Examples
    --------
    >>> from qubovert import PUSO
    >>> H = PUSO()
    >>> H[('a',)] += 1
    >>> H, H.mapping, H.reverse_mapping
    {('a',): 1}, {'a': 0}, {0: 'a'}
    >>> H[('a',)] -= 1
    >>> H, H.mapping, H.reverse_mapping
    {}, {'a': 0}, {0: 'a'}
    >>> H.refresh()
    >>> H, H.mapping, H.reverse_mapping
    {}, {}, {}

    """

    def __init__(self, *args, **kwargs):
        """__init__.

        This class deals with PUSOs that have spin labels that do not range
        from 0 to n-1. Note that it is generally more efficient to initialize
        an empty PUSO object and then build the PUSO, rather than
        initialize a PUSO object with an already built dict.

        Parameters
        ----------
        arguments : define a dictionary with ``dict(*args, **kwargs)``.
            The dictionary will be initialized to follow all the convensions of
            the class.

        Examples
        --------
        >>> puso = PUSO()
        >>> puso[('a',)] += 5
        >>> puso[(0, 'a')] -= 2
        >>> puso -= 1.5
        >>> puso
        {('a',): 5, ('a', 0): -2, (): -1.5}

        >>> puso = PUSO({('a',): 5, (0, 'a', 1): -2, (): -1.5})
        >>> puso
        {('a',): 5, ('a', 0, 1): -2, (): -1.5}

        """
        BO.__init__(self, *args, **kwargs)
        PUSOMatrix.__init__(self, *args, **kwargs)

    def _to_puso(self):
        """to_puso.

        Internal helper method.

        Create and return upper triangular degree PUSO representing
        the problem. The labels will be integers from 0 to n-1.

        Return
        ------
        H : qubovert.utils.PUSOMatrix object.
            The upper triangular PUSO matrix, a PUSOMatrix object.
            For most practical purposes, you can use PUSOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.PUSOMatrix)``.

        """
        H = PUSOMatrix()
        for k, v in self.items():
            key = tuple(sorted(self._mapping[i] for i in k))
            H[key] += v
        return H

    def _create_pubo(self):
        """_create_pubo.

        Internal method to create the PUBO object from self.

        Returns
        -------
        P : qubovert.PUBO object.

        """
        P = puso_to_pubo(self)
        P._mapping = self.mapping
        P._reverse_mapping = self.reverse_mapping
        return P

    def to_pubo(self, deg=None, lam=None, pairs=None):
        """to_pubo.

        Create and return upper triangular degree ``deg`` PUBO representing the
        PUSO problem. The labels will be integers from 0 to n-1.

        Parameters
        ----------
        deg : int >= 0 (optional, defaults to None).
            The degree of the final PUBO. If ``deg`` is None, then the degree
            of the output PUBO will be the same as the degree of ``self``,
            ie see ``self.degree``.
        lam : function (optional, defaults to None).
            Note that if ``deg`` is None or ``deg >= self.degree``, then
            ``lam`` is unneccessary and will not be used.
            If ``lam`` is None, the function ``PUBO.default_lam`` will be used.
            ``lam`` is the penalty factor to introduce in order to enforce the
            ancilla constraints. When we reduce the degree of the model, we add
            penalties to the lower order model in order to enforce ancilla
            variable constraints. These constraints will be multiplied by
            ``lam(v)``, where ``v`` is the value associated with the term that
            it is reducing. For example, a term ``(0, 1, 2): 3`` in the higher
            order model may be reduced to a term ``(0, 3): 3`` for the lower
            order model, and then the fact that ``3`` should be the product
            of ``1`` and ``2`` will be enforced with a penalty weight
            ``lam(3)``.
        pairs : set (optional, defaults to None).
            A set of tuples of variable pairs to prioritize pairing together in
            to degree reduction. If a pair in ``pairs`` is found together in
            the PUSO, it will be chosen as a pair to reduce to a single
            ancilla. You should supply this parameter if you have a good idea
            of an efficient way to reduce the degree of the PUSO. If ``pairs``
            is None, then it will be the empty set ``set()``. In other words,
            no variable pairs will be prioritized, and instead variable pairs
            will be chosen to reduce to an ancilla bases solely on frequency
            of occurrance.

        Return
        ------
        P : qubovert.utils.PUBOMatrix object.
            The upper triangular PUBO matrix, a PUBOMatrix object.
            For most practical purposes, you can use PUBOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.PUBOMatrix)``.

        Notes
        -----
        The penalty that we use to enforce the constraints that the ancilla
        variable ``z`` is equal to the product of the two variables that it is
        replacing, ``xy``, is:

            ``0`` if ``z == xy``,
            ``3*lam(v)`` if ``x == y == 0 and z == 1``, and
            ``lam(v)`` else.

        See https://arxiv.org/pdf/1307.8041.pdf equation 6.

        """
        return self._create_pubo().to_pubo(deg, lam, pairs)

    def to_puso(self, deg=None, lam=None, pairs=None):
        """to_puso.

        Create and return upper triangular degree ``deg`` PUSO representing
        the problem. The labels will be integers from 0 to n-1.

        Parameters
        ----------
        deg : int >= 0 (optional, defaults to None).
            The degree of the final PUSO. If ``deg`` is None, then the degree
            of the output PUSO will be the same as the degree of ``self``,
            ie see ``self.degree``.
        lam : function (optional, defaults to None).
            Note that if ``deg`` is None or ``deg >= self.degree``, then
            ``lam`` is unneccessary and will not be used.
            If ``lam`` is None, the function ``PUBO.default_lam`` will be used.
            ``lam`` is the penalty factor to introduce in order to enforce the
            ancilla constraints. When we reduce the degree of the model, we add
            penalties to the lower order model in order to enforce ancilla
            variable constraints. These constraints will be multiplied by
            ``lam(v)``, where ``v`` is the value associated with the term that
            it is reducing. For example, a term ``(0, 1, 2): 3`` in the higher
            order model may be reduced to a term ``(0, 3): 3`` for the lower
            order model, and then the fact that ``3`` should be the product
            of ``1`` and ``2`` will be enforced with a penalty weight
            ``lam(3)``.
        pairs : set (optional, defaults to None).
            A set of tuples of variable pairs to prioritize pairing together in
            to degree reduction. If a pair in ``pairs`` is found together in
            the PUSO, it will be chosen as a pair to reduce to a single
            ancilla. You should supply this parameter if you have a good idea
            of an efficient way to reduce the degree of the PUSO. If ``pairs``
            is None, then it will be the empty set ``set()``. In other words,
            no variable pairs will be prioritized, and instead variable pairs
            will be chosen to reduce to an ancilla bases solely on frequency
            of occurrance.

        Return
        ------
        H : qubovert.utils.PUSOMatrix object.
            The upper triangular PUSO matrix, a PUSOMatrix object.
            For most practical purposes, you can use PUSOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.PUSOMatrix)``.

        Notes
        -----
        See the ``to_pubo`` docstring for more information on the penalties and
        ``lam``.

        """
        if deg is None or deg >= self.degree:
            return self._to_puso()
        return self._create_pubo().to_puso(deg, lam, pairs)

    def to_qubo(self, lam=None, pairs=None):
        """to_qubo.

        Create and return upper triangular QUBO representing the problem.
        The labels will be integers from 0 to n-1. We introduce ancilla
        variables in order to reduce the degree of the PUSO to a QUBO. The
        solution to the PUSO can be read from the solution to the QUBO by
        using the ``convert_solution`` method.

        Parameters
        ----------
        lam : function (optional, defaults to None).
            If ``lam`` is None, the function ``PUBO.default_lam`` will be used.
            ``lam`` is the penalty factor to introduce in order to enforce the
            ancilla constraints. When we reduce the degree of the PUSO to a
            QUBO, we add penalties to the QUBO in order to enforce ancilla
            variable constraints. These constraints will be multiplied by
            ``lam(v)``, where ``v`` is the value associated with the term that
            it is reducing. For example, a term ``(0, 1, 2): 3`` in the Hquso
            may be reduced to a term ``(0, 3): 3`` for the QUSO, and then the
            fact that ``3`` should be the product of ``1`` and ``2`` will be
            enforced with a penalty weight ``lam(3)``.
        pairs : set (optional, defaults to None).
            A set of tuples of variable pairs to prioritize pairing together in
            to degree reduction. If a pair in ``pairs`` is found together in
            the PUSO, it will be chosen as a pair to reduce to a single
            ancilla. You should supply this parameter if you have a good idea
            of an efficient way to reduce the degree of the PUSO. If ``pairs``
            is None, then it will be the empty set ``set()``. In other words,
            no variable pairs will be prioritized, and instead variable pairs
            will be chosen to reduce to an ancilla bases solely on frequency
            of occurrance.

        Return
        ------
        Q : qubovert.utils.QUBOMatrix object.
            The upper triangular QUBO matrix, an QUBOMatrix object.
            For most practical purposes, you can use QUBOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.QUBOMatrix)``.

        """
        return self._create_pubo().to_qubo(lam, pairs)

    def convert_solution(self, solution, spin=True):
        """convert_solution.

        Convert the solution to the integer labeled PUSO to the solution to
        the originally labeled PUSO.

        Parameters
        ----------
        solution : iterable or dict.
            The PUSO, PUSO, QUSO, or QUSO solution output. The PUSO
            solution output is either a list or tuple where indices specify the
            label of the variable and the element specifies whether it's 0 or 1
            for PUSO (or 1 or -1 for QUSO), or it can be a dictionary that
            maps the label of the variable to is value. The QUSO/QUSO
            solution output includes the assignment for the ancilla variables
            used to reduce the degree of the PUSO.
        spin : bool (optional, defaults to True).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        res : dict.
            Maps spin variable labels to their PUSO solutions values
            {1, -1}.

        Example
        -------
        >>> puso = PUSO({('a',): 5, (0, 'a', 1): -2, (): -1.5})
        >>> puso
        {('a',): 5, ('a', 0, 1): -2, (): -1.5}
        >>> H = puso.to_puso()
        >>> H
        {(0,): 5, (0, 1): -2, (): -1.5}
        >>> puso.convert_solution({0: 1, 1: -1, 2: 1})
        {'a': 1, 0: -1, 1: 1}

        In the next example, notice that we introduce ancilla variables to
        represent that ```(0, 1)`` term. See the ``to_quso`` method for more
        info.

        >>> puso = PUSO({('a',): 5, (0, 'a', 1): -2, (): -1.5})
        >>> puso.mapping
        {'a': 0, 0: 1, 1: 2}
        >>> L = puso.to_quso(3)
        >>> L
        {(0,): 5, (0, 3): -2, (): -2.25,
         (1, 2): 3/4, (2, 3): 3/4, (1, 3): 3/4, (1,): 3/2, (2,): 3/2, (3,): 3}
        >>> puso.convert_solution({0: 1, 1: -1, 2: 1, 2: -1})
        {'a': 1, 0: -1, 1: 1}

        Notes
        -----
        We take ignore the ancilla variable assignments when we convert the
        solution. For example if the conversion from PUSO to QUSO introduced
        an ancilla varable ``z = xy`` where ``x`` and ``y`` are variables of
        the PUSO, then ``solution`` must have values for ``x``, ``y``, and
        ``z``. If the QUSO solver found that ``x = 1``, ``y = -1``, and
        ``z = 1``, then the constraint that ``z = xy`` is not satisfied (one
        possible cause for this is if the ``lam`` argument in ``to_quso`` is
        too small). ``convert_solution`` will return that ``x = 1`` and
        ``y = -1`` and ignore the value of ``z``.

        """
        # this works for converting a solution to the pubo, qubo, puso, or
        # quso formulations, since in the to_quso function all ancilla
        # variables are labeled with integers >= self.num_binary_variables.
        return QUSO.convert_solution(self, solution, spin)

    @staticmethod
    def _check_key_valid(key):
        """_check_key_valid.

        Internal method to check if an input key to the dictionary is valid.
        Checks to see if ``key`` is a tuple.

        Parameters
        ----------
        key : anything, but must be a tuple to be valid.

        Returns
        -------
        None.

        Raises
        ------
        KeyError if the key is invalid.

        """
        # override PUSOMatrix._check_key_valid to allow for noninteger keys.
        if not isinstance(key, tuple):
            raise KeyError(
                "Key formatted incorrectly, must be tuple")



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_bilp.py.

Contains the BILP class. See ``help(qubovert.problems.BILP)``.

"""

from qubovert.utils import QUBOMatrix, solution_type, spin_to_boolean
from qubovert.problems import Problem
import numpy as np


__all__ = 'BILP',


class BILP(Problem):
    r"""BILP.

    Class to manage converting Binary Integer Linear Programming problems to
    and from their QUBO and QUSO formluations. Based on the paper "Ising
    formulations of many NP problems", hereforth designated as [Lucas].

    The goal of the BILP problem is to find the minimum value of
    :math:`\mathbf{c} \cdot \mathbf{x}` subject to
    :math:`S \mathbf{x} = \mathbf{b}`. Here :math:`\mathbf{c}`,
    :math:`\mathbf{b}`, and :math:`S` define the problem, and
    :math:`\mathbf{x}`, is a vector of boolean variables.

    BILP inherits some methods and attributes from the Problem class.
    See ``help(qubovert.problems.Problem)``.

    Example
    -------
    >>> from qubovert.problems import BILP
    >>> from any_module import qubo_solver
    >>> # or you can use my bruteforce solver...
    >>> # from qubovert.utils import solve_qubo_bruteforce as qubo_solver
    >>>
    >>> c = [1, 2, -1]
    >>> S = [[1, 0, 0], [0, 1, -1]]
    >>> b = [0, 1]
    >>> problem = BILP(c, S, b)
    >>>
    >>> Q = problem.to_qubo()
    >>> obj, sol = qubo_solver(Q)
    >>> solution = problem.convert_solution(sol)
    >>>
    >>> print(solution)
    [0 1 0]
    >>> print(problem.is_solution_valid(solution))
    True
    >>> print(obj == np.dot(c, solution))
    True

    """

    def __init__(self, c, S, b):
        r"""__init__.

        The goal of the BILP problem is to find the minimum value of
        :math:`\mathbf{c} \cdot \mathbf{x}` subject to
        :math:`S \mathbf{x} = \mathbf{b}`. Here :math:`\mathbf{c}`,
        :math:`\mathbf{b}`, and :math:`S` define the problem, and
        :math:`\mathbf{x}`, is a vector of boolean variables.
        All naming conventions follow the names in the paper [Lucas].

        Parameters
        ----------
        c : list or numpy one-dim array of integers.
            Array representing the :math:`\mathbf{c}` vector. ``c`` should
            have length ``N``.
        S : list of lists or two-dim numpy array of integers.
            Array representing the :math:`S` matrix. ``S`` should be ``m`` by
            ``N``.
        b : list or one-dim numpy array of integers.
            Array representing the :math:`\mathbf{b}` vector. ``b`` should have
            length ``m``.

        Examples
        --------
        >>> c =
        >>> b =
        >>> S =
        >>> problem = BILP(c, S, b)

        """
        self._c, self._S, self._b = np.array(c), np.array(S), np.array(b)
        self._m, self._N = self._S.shape

        if len(self._c.shape) > 1 or len(self._b.shape) > 1:
            raise ValueError("b and c must be one dimensional")
        elif self._c.shape[0] != self._N or self._b.shape[0] != self._m:
            raise ValueError("Incorrect dimensions")

    @property
    def c(self):
        """c.

        A copy of the ``c`` vector. Updating the copy will not
        update the instance set.

        Return
        ------
        c : numpy array.

        """
        return self._c.copy()

    @property
    def S(self):
        """S.

        A copy of the ``S`` matrix. Updating the copy will not
        update the instance set.

        Return
        ------
        S : numpy array.

        """
        return self._S.copy()

    @property
    def b(self):
        """b.

        A copy of the ``b`` vector. Updating the copy will not
        update the instance set.

        Return
        ------
        b : numpy array.

        """
        return self._b.copy()

    @property
    def num_binary_variables(self):
        """num_binary_variables.

        The number of binary variables that the QUBO and QUSO use.

        Return
        ------
        num : integer.
            The number of variables in the QUBO/QUSO formulation.

        """
        return self._N

    def to_qubo(self, A=None, B=1):
        r"""to_qubo.

        Create and return the BILP problem in QUBO form following
        section 3 of [Lucas]. The Q matrix for the QUBO
        will be returned as an uppertriangular dictionary. Thus, the problem
        becomes minimizing :math:`\sum_{i \leq j} x_i x_j Q_{ij}`. ``A`` and
        ``B`` are parameters to enforce constraints.

        Parameters
        ----------
        A: positive float (optional, defaults to None).
            A enforces the constraints. If ``A`` is
            None, then a default value will be chosen as given in section 3 of
            [Lucas].
        B: positive float that is less than A (optional, defaults to 1).
            See section 3 of [Lucas].

        Return
        ------
        Q : qubovert.utils.QUBOMatrix object.
            The upper triangular QUBO matrix, a QUBOMatrix object.
            For most practical purposes, you can use QUBOMatrix in the
            same way as an ordinary dictionary. For more information,
            see help(qubovert.utils.QUBOMatrix).

        """
        # all naming conventions follow the paper listed in the docstring
        if A is None:
            A = B * self._N

        Q = QUBOMatrix()

        # equation 23
        for i in range(self._N):
            Q[(i,)] += B * self._c[i]

        # equation 22
        for j in range(self._m):
            Qtemp = QUBOMatrix()
            Qtemp += self._b[j]
            for i in range(self._N):
                Qtemp[(i,)] -= self._S[j][i]
            Qtemp *= A * Qtemp
            Q += Qtemp

        return Q

    def convert_solution(self, solution, spin=False):
        r"""convert_solution.

        Convert the solution to the QUBO or QUSO to the solution to the BILP
        problem.

        Parameters
        ----------
        solution : iterable or dict.
            The QUBO or QUSO solution output. The QUBO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool (optional, defaults to False).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        res : np.array.
            An array representing the :math:`\mathbf{x}` vector.

        """
        sol_type = solution_type(solution, 'spin' if spin else 'bool')
        if sol_type == 'spin':
            solution = spin_to_boolean(solution)
        return np.array([int(bool(solution[i])) for i in range(self._N)])

    def is_solution_valid(self, solution, spin=False):
        r"""is_solution_valid.

        Returns whether or not the proposed solution satisfies the constraint
        that :math:`S\mathbf{x} = \mathbf{b}`.

        Parameters
        ----------
        solution : iterable or dict.
            solution can be the output of BILP.convert_solution,
            or the  QUBO or QUSO solver output. The QUBO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool (optional, defaults to False).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        valid : boolean.
            True if the proposed solution is valid, else False.

        """
        if not isinstance(solution, np.ndarray):
            solution = self.convert_solution(solution, spin)

        return np.allclose(
            self._S @ np.array([solution]).T,
            np.array([self._b]).T
        )



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_pubomatrix.py.

This file contains the PUBOMatrix object.

"""

from . import (
    DictArithmetic, hash_function,
    pubo_value, solve_pubo_bruteforce
)


__all__ = 'PUBOMatrix',


class PUBOMatrix(DictArithmetic):
    """PUBOMatrix.

    ``PUBOMatrix`` inherits some methods from ``DictArithmetic``, see
    ``help(qubovert.utils.DictArithmetic)``.

    A class to handle PUBO matrices. It is the same thing as a dictionary
    with some methods modified. Note that each key must be a tuple of integers
    >= 0.

    Note that below we only consider keys that are of length 1 or 2, but they
    can in general be arbitrarily long! See ``qubovert.utils.QUBOMatrix``
    for an object that restricts the length to <= 2.

    One method is that values will always default to 0. Consider the following
    example:

    >>> d = PUBOMatrix()
    >>> print(d[(0,)]) # will print 0
    >>> d[(0,)] += 1
    >>> print(d) # will print {(0,): 1}

    Compared to an ordinary dictionary.

    >>> g = dict()
    >>> print(g[(0,)]) # will raise KeyError
    >>> g[(0,)] += 1 # will raise KeyError, since (0,) was never set

    One method of PUBOMatrix is that it will always keep the PUBO
    upper triangular! Consider the following example:

    >>> d = PUBOMatrix()
    >>> d[(1, 0)] += 2
    >>> print(d)
    >>> # will print {(0, 1): 2}

    One method is that if we set an item to 0, it will be removed. Consider
    the following example:

    >>> d = PUBOMatrix()
    >>> d[(0,)] += 1
    >>> d[(0,)] -= 1
    >>> print(d) # will print {}

    One method is that if we initialize PUBOMatrix with a previous dictionary
    it will be reinitialized to ensure that the PUBOMatrix is upper
    triangular and contains no zero values. Consider the following example:

    >>> d = PUBOMatrix({(0, 0): 1, (1, 0): 2, (2, 0): 0, (2, 0, 1): 1})
    >>> print(d) # will print {(0,): 1, (0, 1): 2, (0, 1, 2): 1}

    We also change the update method so that it follows all the conventions.

    >>> d = PUBOMatrix({(0, 0): 1, (0, 1): 2})
    >>> d.update({(0,): 0, (1, 0): 1, (1, 2): -1})
    >>> print(d)  # will print {(0, 1): 1, (1, 2): -1}

    We also include arithmetic, addition, subtraction, scalar division,
    multiplication, and all those in place. For example,

    >>> d = PUBOMatrix({(0, 0): 1, (0, 1): -2})
    >>> g = d + {(0,): -1}
    >>> print(g) # will print {(0, 1): -2}
    >>> g *= 4
    >>> print(g) # will print {(0, 1): -8}
    >>> g -= {(0, 1): -8}
    >>> print(g) # will print {}

    >>> d = PUBOMatrix({(0, 0): 1, (0, 1): -1})
    >>> g = {(0,): -1, (2,): 1}
    >>> d *= g
    >>> print(d)
    {(0,): -1, (0, 2): 1, (0, 1): 1, (0, 1, 2): -1}

    >>> d = PUBOMatrix({(0, 0): 1, (0, 1): -1})
    >>> print(d ** 2 == d * d)
    True

    Adding or subtracting constants will update the () element of the
    dict.

    >>> d = PUBOMatrix()
    >>> d += 5
    >>> print(d)
    {(): 5}

    Finally, if you try to access a key out of order, it will sort the key. Be
    careful with this, it can cause unexpected behavior if you don't know it.
    For example,

    >>> d = PUBOMatrix()
    >>> d[(0, 1)] += 2
    >>> print(d[(1, 0)])  # will print 2

    >>> d = PUBOMatrix()
    >>> d[(0, 0)] += 2
    >>> print(d[(0,)])  # will print 2

    >>> d = PUBOMatrix()
    >>> d[(0, 0, 3, 2, 2)] += 2
    >>> print(d)  # will print {(0, 2, 3): 2}
    >>> print(d[(0, 3, 2)])  # will print 2

    """

    def __init__(self, *args, **kwargs):
        """__init__.

        This class deals with Binary Optimization matrices. See child classes
        or ``qubovert.utils.DictArithmetic`` for details on the inputs.

        Parameters
        ----------
        arguments : parameters.
            Defined in child classes or in ``qubovert.utils.DictArithmetic``.

        """
        self._degree, self._variables, self._num_binary_variables = 0, set(), 0
        super().__init__(*args, **kwargs)

    def refresh(self):
        """refresh.

        For efficiency, the internal variables for ``degree``,
        ``num_binary_variables``, ``max_index`` are computed as the dictionary
        is being built (and in subclasses such as ``qubovert.PUBO``, properties
        such as ``mapping`` and ``reverse_mapping``). This can cause these
        values to be wrong for some specific situations. Calling ``refresh``
        will rebuild the dictionary, resetting all of the values.

        Examples
        --------
        >>> from qubovert.utils import PUBOMatrix
        >>> P = PUBOMatrix()
        >>> P[(0,)] += 1
        >>> P, P.degree, P.num_binary_variables
        {(0,): 1}, 1, 1
        >>> P[(0,)] -= 1
        >>> P, P.degree, P.num_binary_variables
        {}, 1, 1
        >>> P.refresh()
        >>> P, P.degree, P.num_binary_variables
        {}, 0, 0

        >>> from qubovert import PUBO
        >>> P = PUBO()
        >>> P[('a',)] += 1
        >>> P, P.mapping, P.reverse_mapping
        {('a',): 1}, {'a': 0}, {0: 'a'}
        >>> P[('a',)] -= 1
        >>> P, P.mapping, P.reverse_mapping
        {}, {'a': 0}, {0: 'a'}
        >>> P.refresh()
        >>> P, P.mapping, P.reverse_mapping
        {}, {}, {}

        """
        d = self.copy()
        super().clear()
        self.__init__(d)

    def clear(self):
        """clear.

        For efficiency, the internal variables for ``degree``,
        ``num_binary_variables``, ``max_index`` are computed as the dictionary
        is being built (and in subclasses such as ``qubovert.PUBO``, properties
        such as ``mapping`` and ``reverse_mapping``). This can cause these
        values to be wrong for some specific situations. Thus, when we clear,
        we also need to reset all of these cached values. This function
        remove all the elments from ``self`` and resets the cached values.

        """
        super().clear()
        self.__init__()

    @property
    def degree(self):
        """degree.

        Return the degree of the problem.

        Return
        ------
        deg : int.

        """
        return self._degree

    @property
    def variables(self):
        """variables.

        Return a set of all the variables in the dict.

        Returns
        -------
        res : set.

        """
        return self._variables.copy()

    @property
    def offset(self):
        """offset.

        Get the part that does not depend on any variables. Ie the value
        corresponding to the () key.

        Returns
        -------
        offset : float.

        """
        return self[()]

    @property
    def num_binary_variables(self):
        """num_binary_variables.

        Return the number of binary variables in the problem.

        Return
        ------
        n : int.
            Number of binary variables in the problem.

        """
        return self._num_binary_variables

    @property
    def max_index(self):
        """max_index.

        Returns the maximum label index of the problem. If the problem is
        labeled with integers from 0 to ``n-1``, then ``max_index`` will give
        the same result as ``num_binary_variables - 1``.

        Return
        ------
        n : int or None.
            Max label index of the problem dictionary. If the dict is empty,
            then this returns None.

        """
        return max(self._variables) if self._variables else None

    @classmethod
    def squash_key(cls, key):
        """squash_key.

        Will convert the input key into the standard form for PUBOMatrix /
        QUBOMatrix. It will get rid of duplicates and sort. This method
        will check to see if the input key is valid.

        Parameters
        ----------
        key : tuple of integers.

        Return
        ------
        k : tuple of integers.
            A sorted and squashed version of ``key``.

        Raises
        ------
        KeyError if the key is invalid.

        Example
        -------
        >>> squash_key((0, 4, 0, 3, 3, 2))
        >>> (0, 2, 3, 4)

        """
        # if f is not None, then it is the squashed key (see QUBOMatrix)
        f = cls._check_key_valid(key)
        # here we use hash because some other classes that are subclasses of
        # this class will allow elements of the key to be strings! So we want
        # to still have something consistent to sort by. But for this class,
        # it doesn't make a difference, because hash_funcition(i) == i when i
        # is an int.
        return f or tuple(sorted(set(key), key=lambda x: hash_function(x)))

    @staticmethod
    def _check_key_valid(key):
        """_check_key_valid.

        Internal method to check if an input key to the dictionary is valid.
        Checks to see if ``key`` is a tuple of non negative integers.

        Parameters
        ----------
        key : anything, but must be a tuple to be valid.

        Returns
        -------
        None.

        Raises
        ------
        KeyError if the key is invalid.

        """
        incorrect_format = (
            not isinstance(key, tuple) or
            any(not isinstance(k, int) or k < 0 for k in key)
        )

        if incorrect_format:
            raise KeyError(
                "Key formatted incorrectly, must be tuple of non negative "
                "integers")

    def __getitem__(self, key):
        """__getitem__.

        Overrides the dict.__getitem__ command so that a KeyError is not
        thrown if `key` is not in the dictionary. The key will also be
        reformatted according to ``squash_key`` before indexing.

        Parameters
        ----------
        key : tuple of integers.
            Element of the dictionary.

        Return
        ------
        value : numeric
            the value corresponding to the key if the key is in the dictionary,
            otherwise returns 0.

        """
        return super().__getitem__(self.__class__.squash_key(key))

    def __setitem__(self, key, value):
        """__setitem__.

        Overrides the dict.__setitem__ command. If `value` is equal to 0, then
        the key will be removed from the dictionary. Thus no elements will ever
        have zero value. Additionally, this method will keep self upper
        triangular, so if key[0] > key[1], then we will call
        ``__setitem__((key[1], key[0]), value)``. Finally, keys
        will be squashed, see ``squash_keys``.

        Parameters
        ----------
        key : tuple of integers.
            Element of the dictionary.
        value : numeric.
            Value corresponding to the key.

        """
        k = self.__class__.squash_key(key)
        if value:
            self._degree = max(self._degree, len(k))
            for i in filter(lambda x: x not in self._variables, k):
                self._variables.add(i)
                self._num_binary_variables += 1
        super().__setitem__(k, value)

    def is_solution_valid(self, solution):
        """is_solution_valid.

        Included for consistency with other problem classes. Always returns
        True since this is an unconstrainted problem.

        Parameters
        ----------
        solution : iterable or dict.

        Return
        ------
        valid : bool.
            Always returns True.

        """
        return True

    def solve_bruteforce(self, all_solutions=False):
        """solve_bruteforce.

        Solve the problem bruteforce. THIS SHOULD NOT BE USED FOR LARGE
        PROBLEMS! This is the exact same as calling
        ``qubovert.utils.solve_pubo_bruteforce(self, all_solutions)[1]``.

        Parameters
        ----------
        all_solutions : bool.
            See the description of the ``all_solutions`` parameter in
            ``qubovert.utils.solve_pubo_bruteforce``.

        Return
        ------
        res : the second element of the two element tuple that is returned from
            ``qubovert.utils.solve_pubo_bruteforce``.

        """
        return solve_pubo_bruteforce(self,
                                     all_solutions, self.is_solution_valid)[1]

    def value(self, x):
        r"""value.

        Find the value of
        :math:`\sum_{i,...,j} P_{i...j} x_{i} ... x_{j}`. Calling
        ``self.value(x)`` is the same as calling
        ``qubovert.utils.pubo_value(x, self)``.

        Parameters
        ----------
        x : dict or iterable.
            Maps boolean variable indices to their boolean values, 0 or 1. Ie
            ``x[i]`` must be the boolean value of variable i.

        Return
        ------
        value : float.
            The value of the PUBO with the given assignment `x`. Ie

        Example
        -------
        >>> from qubovert.utils import QUBOMatrix, PUBOMatrix
        >>> from qubovert import QUBO, PUBO

        >>> P = PUBOMatrix({(0, 0): 1, (0, 1): -1})
        >>> x = {0: 1, 1: 0}
        >>> P.value(x)
        1

        >>> Q = QUBOMatrix({(0, 0): 1, (0, 1): -1})
        >>> x = {0: 1, 1: 0}
        >>> Q.value(x)
        1

        >>> P = PUBO({(0, 0): 1, (0, 1): -1})
        >>> x = {0: 1, 1: 0}
        >>> P.value(x)
        1

        >>> Q = QUBO({(0, 0): 1, (0, 1): -1})
        >>> x = {0: 1, 1: 0}
        >>> Q.value(x)
        1

        """
        return pubo_value(x, self)



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_pcbo.py.

Contains the PCBO class. See ``help(qubovert.PCBO)``.

"""

from . import PUBO
from .utils import QUBOVertWarning, num_bits, approximate_pubo_extrema
from .sat import OR, XOR, BUFFER, NOT, AND


__all__ = 'PCBO', 'boolean_var', 'integer_var'


# special constraint forms

def _special_constraints_eq_zero(pcbo, P, lam):
    """_special_constraints_eq_zero.

    See if the constraint that ``P == 0`` matches any special forms.

    Parameters
    ----------
    pcbo : PCBO object.
        The PCBO to add the constraint to.
    P : PUBO object.
        The PUBO constraint such that ``P == 0``.
        are present.
    lam : float > 0 or sympy.Symbol.
        Langrange multiplier to penalize violations of the constraint.

    Return
    ------
    success : bool.
        True if a special constraint was found and added to ``pcbo``, else
        False.

    """
    keys, values = tuple(P.keys()), tuple(P.values())

    # if P is of the form z == x * y. ie z == AND(x, y)
    if (
            not P.offset and P.num_binary_variables == 3 and
            P.num_terms == 2 and values[0] == -values[1] and
            {len(keys[0]), len(keys[1])} == {1, 2}
    ):
        a, = keys[0] if len(keys[0]) == 1 else keys[1]
        b, c = keys[0] if len(keys[0]) == 2 else keys[1]
        pcbo += PCBO().add_constraint_eq_AND(a, b, c, lam=lam)
        return True

    # if P is of the form z == 1 - x * y. ie z == NAND(x, y)

    # if P is of the form z == x * y. ie z == OR(x, y)

    # if P is of the form z == x * y. ie z == NOR(x, y)

    return False


def _special_constraints_le_zero(pcbo, P, lam, log_trick, bounds):
    """_special_constraints_le_zero.

    See if the constraint that ``P <= 0`` matches any special forms.

    Parameters
    ----------
    pcbo : PCBO object.
        The PCBO to add the constraint to.
    P : PUBO object.
        The PUBO constraint such that ``P <= 0``.
        are present.
    lam : float > 0 or sympy.Symbol.
        Langrange multiplier to penalize violations of the constraint.
    log_trick : bool.
        Whether or not to use the log trick to enforce the inequality
        constraint.
    bounds : two element tuple.
        A tuple ``(min, max)``, the minimum and maximum values that the
        PUBO ``P`` can take.

    Return
    ------
    success : bool.
        True if a special constraint was found and added to ``pcbo``, else
        False.

    """
    min_val, max_val = bounds

    # P without offset defined for convenience
    P_wo_offset = P - P.offset

    # if P is of the form sum(x_i) <= 1.
    if P.offset == -1 and all(x == 1 for x in P_wo_offset.values()):
        pcbo += lam * P * P_wo_offset / 2
        return True

    # if P is of the form P_wo_offset <= -P.offset and P_wo_offset is always
    # >= 0, and (important!) log_trick is False!!
    elif (
            not log_trick and not min_val - P.offset and
            P.offset <= 0 and min_val
    ):
        # We have that P_wo_offset <= P.offset and min(P_wo_offset) = 0. So we
        # can do a penalty lam(P_wo_offset - sum(ancillas))**2

        # create ancillas
        ancillas = PUBO()
        for i in range(num_bits(-P.offset, log_trick)):
            ancillas[(pcbo._next_ancilla,)] += 1

        diff = P_wo_offset - ancillas
        pcbo += lam * diff * diff

        return True

    # if P is of the form 1 <= x + y, then it is the same as OR(x, y)
    elif (P.offset == 1 and len(P_wo_offset) == 2 and
          set(P_wo_offset.values()) == {-1}):
        variables = tuple(P_wo_offset.keys())
        x, y = AND(*variables[0]), AND(*variables[1])
        pcbo += PCBO().add_constraint_OR(x, y, lam=lam)
        return True

    # if P is of the form x <= y
    elif not P.offset and len(P) == 2 and set(P.values()) == {1, -1}:
        coef = {v: k for k, v in P.items()}
        x, y = AND(*coef[1]), AND(*coef[-1])
        pcbo += lam * x * (1 - y)
        return True

    return False


# helpers

def _get_bounds(P, bounds):
    """_get_bounds.

    Compute the missing bounds of ``P``.

    Parameters
    ----------
    P : PUBO object.
    bounds : two element tuple or None.
        A tuple ``(min, max)``, the minimum and maximum values that the
        PUBO ``P`` can take. If ``bounds`` is None, then they will be
        calculated (approximately), or if either of the elements of
        ``bounds`` is None, then that element will be calculated
        (approximately).

    Return
    ------
    res : two element tuple.
        The bounds of ``P``.

    """
    if bounds is None or bounds == (None, None):
        bounds = approximate_pubo_extrema(P)
    elif bounds[0] is None:
        bounds = approximate_pubo_extrema(P)[0], bounds[1]
    elif bounds[1] is None:
        bounds = bounds[0], approximate_pubo_extrema(P)[1]

    return bounds


def boolean_var(name):
    """boolean_var.

    Create a PCBO (see ``qubovert.PCBO``) from a single boolean variable.

    Parameters
    ----------
    name : any hashable object.
        Name of the boolean variable.

    Return
    ------
    pcbo : qubovert.PCBO object.
        The model representing the boolean variable.

    Examples
    --------
    >>> from qubovert import boolean_var, PCBO
    >>>
    >>> x0 = boolean_var("x0")
    >>> print(x0)
    {('x0',): 1}
    >>> print(isinstance(x0, PCBO))
    True
    >>> print(x0.name)
    x0

    >>> x = [boolean_var('x{}'.format(i)) for i in range(5)]
    >>> pcbo = sum(x)
    >>> print(pcbo)
    {('x0',): 1, ('x1',): 1, ('x2',): 1, ('x3',): 1, ('x4',): 1}
    >>> pcbo **= 2
    >>> print(pcbo)
    {('x0',): 1, ('x0', 'x1'): 2, ('x2', 'x0'): 2, ('x3', 'x0'): 2,
     ('x4', 'x0'): 2, ('x1',): 1, ('x2', 'x1'): 2, ('x3', 'x1'): 2,
     ('x4', 'x1'): 2, ('x2',): 1, ('x2', 'x3'): 2, ('x2', 'x4'): 2, ('x3',): 1,
     ('x4', 'x3'): 2, ('x4',): 1}
    >>> pcbo *= -1
    >>> print(pcbo.solve_bruteforce())
    {'x0': 1, 'x1': 1, 'x2': 1, 'x3': 1, 'x4': 1}
    >>> pcbo.add_constraint_eq_zero(x[0] + x[1])
    >>> print(pcbo.solve_bruteforce())
    {'x0': 0, 'x1': 0, 'x2': 1, 'x3': 1, 'x4': 1}

    """
    pcbo = PCBO({(name,): 1})
    pcbo.name = name
    return pcbo


def integer_var(prefix, num_bits, log_trick=True):
    """integer_var.

    Return a PCBO object representing an integer variable with `num_bits`
    bits.

    Parameters
    ----------
    prefix : str.
        The prefix for the boolean variable names.
    num_bits : int.
        Number of bits to represent the integer variable with.
    log_trick : bool (optional, defaults to True).
        Whether or not to use a log encoding for the integer.

    Return
    ------
    i : qubovert.PCBO object.

    Example
    -------
    >>> from qubovert import integer_var
    >>> var = integer_var('a', 4)
    >>> print(var)
    {('a0',): 1, ('a1',): 2, ('a2',): 4, ('a3',): 8}
    >>> print(var.name)
    a

    >>> from qubovert import integer_var
    >>> var = integer_var('a', 4, log_trick=False)
    >>> print(var)
    {('a0',): 1, ('a1',): 1, ('a2',): 1, ('a3',): 1}

    """
    var = PCBO()
    var.name = prefix
    for i in range(num_bits):
        var[(str(prefix) + str(i),)] = pow(2, i) if log_trick else 1
    return var


# main class

class PCBO(PUBO):
    """PCBO.

    This class deals with Polynomial Constrained Boolean Optimization problems.
    PCBO inherits some methods and attributes from the ``PUBO`` class. See
    ``help(qubovert.PUBO)``.

    ``PCBO`` has all the same methods as ``PUBO``, but adds some constraint
    methods; namely

    - ``add_constraint_eq_zero(P, lam=1, ...)`` enforces that ``P == 0`` by
      penalizing with ``lam``,
    - ``add_constraint_ne_zero(P, lam=1, ...)`` enforces that ``P != 0`` by
      penalizing with ``lam``,
    - ``add_constraint_lt_zero(P, lam=1, ...)`` enforces that ``P < 0`` by
      penalizing with ``lam``,
    - ``add_constraint_le_zero(P, lam=1, ...)`` enforces that ``P <= 0`` by
      penalizing with ``lam``,
    - ``add_constraint_gt_zero(P, lam=1, ...)`` enforces that ``P > 0`` by
      penalizing with ``lam``, and
    - ``add_constraint_ge_zero(P, lam=1, ...)`` enforces that ``P >= 0`` by
      penalizing with ``lam``.

    Each of these takes in a PUBO ``P`` and a lagrange multiplier ``lam`` that
    defaults to 1. See each of their docstrings for important details on their
    implementation.

    We then implement logical operations:

    - ``add_constraint_AND``, ``add_constraint_NAND``,
    - ``add_constraint_eq_AND``, ``add_constraint_eq_NAND``,
    - ``add_constraint_OR``, ``add_constraint_NOR```,
    - ``add_constraint_eq_OR``, ``add_constraint_eq_NOR``,
    - ``add_constraint_XOR``, ``add_constraint_XNOR``,
    - ``add_constraint_eq_XOR``, ``add_constraint_eq_XNOR``,
    - ``add_constraint_BUFFER``, ``add_constraint_NOT``,
    - ``add_constraint_eq_BUFFER``, ``add_constraint_eq_NOT``.

    See each of their docstrings for important details on their implementation.

    Notes
    -----
    - Variables names that begin with ``"__a"`` should not be used since they
      are used internally to deal with some ancilla variables to enforce
      constraints.
    - The ``self.solve_bruteforce`` method will solve the PCBO ensuring that
      all the inputted constraints are satisfied. Whereas
      ``qubovert.utils.solve_pubo_bruteforce(self)`` or
      ``qubovert.utils.solve_pubo_bruteforce(self.to_pubo())`` will solve the
      PUBO created from the PCBO. If the inputted constraints are not enforced
      strong enough (ie too small lagrange multipliers) then these may not give
      the correct result, whereas ``self.solve_bruteforce()`` will always give
      the correct result (ie one that satisfies all the constraints).

    Examples
    --------
    See ``qubovert.PUBO`` for more examples of using PCBO without constraints.

    >>> H = PCBO()
    >>> H.add_constraint_eq_zero({('a', 1): 2, (1, 2): -1, (): -1})
    >>> H
    {('a', 1, 2): -4, (1, 2): 3, (): 1}
    >>> H -= 1
    >>> H
    {('a', 1, 2): -4, (1, 2): 3}

    >>> H = PCBO()
    >>>
    >>> # minimize -x_0 - x_1 - x_2
    >>> for i in (0, 1, 2):
    >>>     H[(i,)] -= 1
    >>>
    >>> # subject to constraints
    >>> H.add_constraint_eq_zero(  # enforce that x_0 x_1 - x_2 == 0
    >>>     {(0, 1): 1, (2,): -1}
    >>> ).add_constraint_lt_zero(  # enforce that x_1 x_2 + x_0 < 1
    >>>     {(1, 2): 1, (0,): 1, (): -1}
    >>> )
    >>> print(H)
    >>> # {(1,): -2, (2,): -1, (0, 1): 2, (1, 2): 2, (0, 1, 2): 2}
    >>>
    >>> print(H.solve_bruteforce(all_solutions=True))
    >>> # [{0: 0, 1: 1, 2: 0}]
    >>>
    >>> Q = H.to_qubo()
    >>> solutions = [H.convert_solution(sol)
    >>>              for sol in Q.solve_bruteforce(all_solutions=True)]
    >>> print(solutions)
    >>> # [{0: 0, 1: 1, 2: 0}]  # matches the PCBO solution!
    >>>
    >>> L = H.to_quso()
    >>> solutions = [H.convert_solution(sol)
    >>>              for sol in L.solve_bruteforce(all_solutions=True)]
    >>> print(solutions)
    >>> # [{0: 0, 1: 1, 2: 0}]  # matches the PCBO solution!

    Enforce that c == a b

    >>> H = PCBO().add_constraint_eq_AND('c', 'a', 'b')
    >>> H
    {('c',): 3, ('b', 'a'): 1, ('c', 'a'): -2, ('c', 'b'): -2}

    >>> from any_module import qubo_solver
    >>> # or from qubovert.utils import solve_qubo_bruteforce as qubo_solver
    >>> H = PCBO()
    >>>
    >>> # make it favorable to AND variables a and b, and variables b and c
    >>> H.add_constraint_AND('a', 'b').add_constraint_AND('b', 'c')
    >>>
    >>> # make it favorable to OR variables b and c
    >>> H.add_constraint_OR('b', 'c')
    >>>
    >>> # make it favorable to (a AND b) OR (c AND d)
    >>> H.add_constraint_OR(['a', 'b'], ['c', 'd'])
    >>>
    >>> H
    {('b', 'a'): -2, (): 4, ('b',): -1, ('c',): -1,
     ('c', 'd'): -1, ('c', 'd', 'b', 'a'): 1}
    >>> Q = H.to_qubo()
    >>> Q
    {(): 4, (0,): -1, (2,): -1, (2, 3): 1, (4,): 6, (0, 4): -4,
     (1, 4): -4, (5,): 6, (2, 5): -4, (3, 5): -4, (4, 5): 1}
    >>> obj_value, sol = qubo_solver(Q)
    >>> sol
    {0: 1, 1: 1, 2: 1, 3: 0, 4: 1, 5: 0}
    >>> solution = H.convert_solution(sol)
    >>> solution
    {'b': 1, 'a': 1, 'c': 1, 'd': 0}

    """

    def __init__(self, *args, **kwargs):
        """__init__.

        This class deals with polynomial constrained boolean optimization.
        Note that it is generally more efficient to initialize an empty PCBO
        object and then build the PCBO, rather than initialize a PCBO object
        with an already built dict.

        Parameters
        ----------
        arguments : define a dictionary with ``dict(*args, **kwargs)``.
            The dictionary will be initialized to follow all the convensions of
            the class. Alternatively, ``args[0]`` can be a PCBO object.

        Examples
        --------
        >>> pcbo = PCBO()
        >>> pcbo[('a',)] += 5
        >>> pcbo[(0, 'a')] -= 2
        >>> pcbo -= 1.5
        >>> pcbo
        {('a',): 5, ('a', 0): -2, (): -1.5}
        >>> pcbo.add_constraint_eq_zero({('a',): 1}, lam=5)
        >>> pcbo
        {('a',): 10, ('a', 0): -2, (): -1.5}

        >>> pcbo = PCBO({('a',): 5, (0, 'a', 1): -2, (): -1.5})
        >>> pcbo
        {('a',): 5, ('a', 0, 1): -2, (): -1.5}

        """
        # use self.__class__ here because PCSO uses this code as well.
        super(self.__class__, self).__init__(*args, **kwargs)
        if len(args) == 1 and isinstance(args[0], self.__class__):
            self._constraints = args[0].constraints
            self._ancilla = args[0].num_ancillas
        else:
            self._ancilla, self._constraints = 0, {}

    def update(self, *args, **kwargs):
        """update.

        Update the PCBO but following all the conventions of this class.

        Parameters
        ----------
        arguments : defines a dictionary or PCBO.
            Ie ``d = dict(*args, **kwargs)``.
            Each element in d will be added in place to this instance following
            all the required convensions.

        """
        # use self.__class__ here because PCSO uses this code as well.
        super(self.__class__, self).update(*args, **kwargs)
        if len(args) == 1 and isinstance(args[0], self.__class__):
            for k, v in args[0]._constraints.items():
                self._constraints.setdefault(k, []).extend(v)

    @property
    def constraints(self):
        """constraints.

        Return the constraints of the PCBO.

        Return
        ------
        res : dict.
            The keys of ``res`` are some or all of
            ``'eq'``, ``'ne'``, ``'lt'``, ``'le'``, ``'gt'``, and ``'ge'``.
            The values are lists of ``qubovert.PUBO`` objects. For a
            given key, value pair ``k, v``, the ``v[i]`` element represents
            the PUBO ``v[i]`` being == 0 if ``k == 'eq'``,
            != 0 if ``k == 'ne'``,
            < 0 if ``k == 'lt'``, <= 0 if ``k == 'le'``,
            > 0 if ``k == 'gt'``, >= 0 if ``k == 'ge'``.

        """
        return {k: [x.copy() for x in v] for k, v in self._constraints.items()}

    def _append_constraint(self, key, constraint):
        """_append_constraint.

        Internal method to add a constraint to the constraints dictionary.

        Parameters
        ----------
        key : str.
            One of ``'eq'``, ``'ne'``, ``'lt'``, ``'le'``, ``'gt'``, or
            ``'ge'``.
        constraint : qubovert.PUBO object.

        """
        self._constraints.setdefault(key, []).append(constraint)

    def _pop_constraint(self, key):
        """_append_constraint.

        Internal method to remove the most recently added constraint.

        Parameters
        ----------
        key : str.
            One of ``'eq'``, ``'ne'``, ``'lt'``, ``'le'``, ``'gt'``, or
            ``'ge'``.

        """
        if self._constraints.get(key, []):
            self._constraints[key].pop()
            if not self._constraints[key]:
                self._constraints.pop(key)

    @property
    def num_ancillas(self):
        """num_ancillas.

        Return the number of ancilla variables introduced to the PCBO in
        order to enforce the inputted constraints.

        Returns
        -------
        num : int.
            Number of ancillas in the PCBO.

        """
        return self._ancilla

    @property
    def _next_ancilla(self):
        """_next_ancilla.

        Get the next available ancilla bit and increment.

        Return
        ------
        a : int.
            Ancilla bit label.

        """
        self._ancilla += 1
        return "__a%d" % (self._ancilla - 1)

    @classmethod
    def remove_ancilla_from_solution(cls, solution):
        """remove_ancilla_from_solution.

        Take a solution to the PCBO and remove all the ancilla variables, (
        represented by `_a` prefixes).

        Parameters
        ----------
        solution : dict.
            Must be the solution in terms of the original variables. Thus if
            ``solution`` is the solution to the ``self.to_pubo``,
            ``self.to_qubo``, ``self.to_puso``, or ``self.to_quso``
            formulations, then you should first call ``self.convert_solution``.
            See ``help(self.convert_solution)``.

        Return
        ------
        res : dict.
            The same as ``solution`` but with all the ancilla bits removed.

        """
        return {k: v for k, v in solution.items() if str(k)[:3] != "__a"}

    def is_solution_valid(self, solution):
        """is_solution_valid.

        Finds whether or not the given solution satisfies the constraints.

        Parameters
        ----------
        solution : dict.
            Must be the solution in terms of the original variables. Thus if
            ``solution`` is the solution to the ``self.to_pubo``,
            ``self.to_qubo``, ``self.to_puso``, or ``self.to_quso``
            formulations, then you should first call ``self.convert_solution``.
            See ``help(self.convert_solution)``.

        Return
        ------
        valid : bool.
            Whether or not the given solution satisfies the constraints.

        """
        if any(v.value(solution) != 0
               for v in self._constraints.get('eq', [])):
            return False

        if any(v.value(solution) == 0
               for v in self._constraints.get('ne', [])):
            return False

        if any(v.value(solution) >= 0
               for v in self._constraints.get("lt", [])):
            return False

        if any(v.value(solution) > 0
               for v in self._constraints.get("le", [])):
            return False

        if any(v.value(solution) <= 0
               for v in self._constraints.get("gt", [])):
            return False

        if any(v.value(solution) < 0
               for v in self._constraints.get("ge", [])):
            return False

        return True

    # override
    def __round__(self, ndigits=None):
        """round.

        Round values of the PCBO object.

        Parameters
        ----------
        ndigits : int.
            Number of decimal digits to round to.

        Returns
        -------
        res : PCBO object.
            Copy of self but with each value rounded to ``ndigits`` decimal
            digits. Each value has a type according to the docstring
            specifications of ``round``, see ``help(round)``.

        """
        # use self.__class__ here because PCSO uses this code as well.
        d = super(self.__class__, self).__round__(ndigits)
        d._constraints = self.constraints
        return d

    # override
    def subs(self, *args, **kwargs):
        """subs.

        Replace any ``sympy`` symbols that are used in the dict with values.
        Please see ``help(sympy.Symbol.subs)`` for more info.

        Parameters
        ----------
        arguments : substitutions.
            Same parameters as are inputted into ``sympy.Symbol.subs``.

        Returns
        -------
        res : PCBO object.
            Same as ``self`` but with all the symbols replaced with values.

        """
        # use self.__class__ here because PCSO uses this code as well.
        d = super(self.__class__, self).subs(*args, **kwargs)
        d._constraints = {
            k: [P.subs(*args, **kwargs) for P in v]
            for k, v in self._constraints.items()
        }
        return d

    def add_constraint_eq_zero(self,
                               P, lam=1,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_eq_zero.

        Enforce that ``P == 0`` by penalizing invalid solutions with ``lam``.

        Parameters
        ----------
        P : dict representing a PUBO.
            The PUBO constraint such that P == 0. Note that ``P`` will be
            converted to a ``qubovert.PUBO`` object if it is not already, thus
            it must follow the conventions, see ``help(qubovert.PUBO)``.
            Please note that if ``P`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUBO ``P`` can take. If ``bounds`` is None, then they may be
            calculated (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        The following enforces that :math:`\prod_{i=0}^{3} x_i == 0`.

        >>> H = PCBO()
        >>> H.add_constraint_eq_zero({(0, 1, 2, 3): 1})
        >>> H
        {(0, 1, 2, 3): 1}

        The following enforces that :math:`\sum_{i=1}^{3} i x_i x_{i+1} == 0`.

        >>> H = PCBO()
        >>> H.add_constraint_eq_zero({(1, 2): 1, (2, 3): 2, (3, 4): 3})
        >>> H
        {(1, 2): 1, (1, 2, 3): 4, (1, 2, 3, 4): 6,
         (2, 3): 4, (2, 3, 4): 12, (3, 4): 9}

        Here we show how operations can be strung together.

        >>> H = PCBO()
        >>> H.add_constraint_eq_zero(
                {(0, 1): 1}
            ).add_constraint_eq_zero(
                {(1, 2): 1, (): -1}
            )
        >>> H
        {(0, 1): 1, (1, 2): -1, (): 1}

        """
        P = PUBO(P)
        self._append_constraint("eq", P)

        if _special_constraints_eq_zero(self, P, lam):
            return self

        min_val, max_val = _get_bounds(P, bounds)

        if min_val == max_val == 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint is always satisfied")
        elif min_val > 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint cannot be satisfied")
            self += lam * P
        elif max_val < 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint cannot be satisfied")
            self -= lam * P
        elif min_val == 0:
            self += lam * P
        elif max_val == 0:
            self -= lam * P
        else:
            self += lam * P * P

        return self

    def add_constraint_ne_zero(self,
                               P, lam=1, log_trick=True,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_ne_zero.

        Enforce that ``P != 0`` by penalizing invalid solutions with ``lam``.
        See Notes below for more details.

        Parameters
        ----------
        P : dict representing a PUBO.
            The PUBO constraint such that P != 0. Note that ``P`` will be
            converted to a ``qubovert.PUBO`` object if it is not already, thus
            it must follow the conventions, see ``help(qubovert.PUBO)``.
            Please note that if ``P`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        log_trick : bool (optional, defaults to True).
            Whether or not to use the log trick to enforce the inequality
            constraint. See Notes below for more details.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUBO ``P`` can take. If ``bounds`` is None, then they will be
            calculated (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Notes
        -----
        - There is no general way to enforce non integer inequality
          constraints. Thus this function is only guarenteed to work for
          integer inequality constraints (ie constraints of the form
          :math:`x_0 + 2x_1 + ... != 0`). However, it can be used for non
          integer inequality constraints, but it is recommended that the value
          of ``lam`` be set small, since valid solutions may still recieve a
          penalty to the objective function.
        - To enforce the inequality constraint, ancilla bits will be
          introduced (labels with `_a`). If ``log_trick`` is ``True``, then
          approximately :math:`\log_2 |\min_x \text{P.value(x)}|`
          ancilla bits will be used. If ``log_trick`` is ``False``, then
          approximately :math:`|\min_x \text{P.value(x)}|` ancilla
          bits will be used.

        Examples
        --------
        Enforce that :math:`-x_a x_b x_c + x_a -4x_a x_b + 3x_c != 2`.

        >>> H = PCBO().add_constraint_ne_zero(
                {('a', 'b', 'c'): -1, ('a',): 2,
                ('a', 'b'): -4, ('c',): 3, (): -2}
            )
        >>> print(H)
        {('c', 'a', 'b'): -19, ('__a0', 'c', 'a', 'b'): -4,
         ('__a0', 'c', '__a1', 'a', 'b'): -4, ('c', '__a1', 'a', 'b'): 2,
         ('__a0', 'c', '__a2', 'a', 'b'): -8, ('c', '__a2', 'a', 'b'): 4,
         ('__a0', '__a3', 'c', 'a', 'b'): -16, ('__a3', 'c', 'a', 'b'): 8,
         ('__a0', 'c', '__a4', 'a', 'b'): -32, ('c', '__a4', 'a', 'b'): 16,
         ('a',): -8, ('c', 'a'): 12, ('__a0', 'a'): 8,
         ('__a0', '__a1', 'a'): 8, ('__a1', 'a'): -4,
         ('__a0', '__a2', 'a'): 16, ('__a2', 'a'): -8,
         ('__a0', '__a3', 'a'): 32, ('__a3', 'a'): -16,
         ('__a0', '__a4', 'a'): 64, ('__a4', 'a'): -32, ('a', 'b'): 24,
         ('__a0', 'a', 'b'): -16, ('__a0', '__a1', 'a', 'b'): -16,
         ('__a1', 'a', 'b'): 8, ('__a0', '__a2', 'a', 'b'): -32,
         ('__a2', 'a', 'b'): 16, ('__a0', '__a3', 'a', 'b'): -64,
         ('__a3', 'a', 'b'): 32, ('__a0', '__a4', 'a', 'b'): -128,
         ('__a4', 'a', 'b'): 64, ('__a0', 'c'): 12, ('__a0', 'c', '__a1'): 12,
         ('c', '__a1'): -6, ('__a0', 'c', '__a2'): 24, ('c', '__a2'): -12,
         ('__a0', '__a3', 'c'): 48, ('__a3', 'c'): -24,
         ('__a0', 'c', '__a4'): 96, ('c', '__a4'): -48, ('c',): -9, (): 9,
         ('__a0',): -8, ('__a0', '__a1'): -8, ('__a1',): 7,
         ('__a0', '__a2'): -16, ('__a2',): 16, ('__a3',): 40,
         ('__a0', '__a4'): -64, ('__a4',): 112, ('__a2', '__a1'): 4,
         ('__a3', '__a1'): 8, ('__a4', '__a1'): 16, ('__a3', '__a2'): 16,
         ('__a4', '__a2'): 32, ('__a0', '__a3'): -32, ('__a3', '__a4'): 64}
        >>> print(H.is_solution_valid({'b': 0, 'c': 0, 'a': 1}))
        False
        >>> print(H.is_solution_valid({'b': 0, 'c': 1, 'a': 1}))
        True

        """
        P = PUBO(P)
        self._append_constraint("ne", P)

        min_val, max_val = _get_bounds(P, bounds)

        if min_val == max_val == 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint cannot be satisfied")
            self += lam
        elif min_val > 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint is always satisfied")
        elif max_val < 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint is always satisfied")
        elif min_val == 0:
            self.add_constraint_gt_zero(
                P, lam=lam,
                bounds=(min_val, max_val), suppress_warnings=suppress_warnings
            )
            self._pop_constraint('gt')
        elif max_val == 0:
            self.add_constraint_lt_zero(
                P, lam=lam,
                bounds=(min_val, max_val), suppress_warnings=suppress_warnings
            )
            self._pop_constraint('lt')

        else:
            # don't mutate the P that we put in self._constraints
            P = P.copy()
            sign = 2 * boolean_var(self._next_ancilla) - 1
            P += sign
            max_val += 1
            min_val -= 1
            for i in range(num_bits(max_val - min_val - 1, log_trick)):
                v = pow(2, i) if log_trick else 1
                P += sign * v * boolean_var(self._next_ancilla)
                max_val += v
                min_val -= v

            self.add_constraint_eq_zero(
                P, lam=lam,
                bounds=(min_val, max_val), suppress_warnings=True
            )
            self._pop_constraint("eq")

        return self

    def add_constraint_lt_zero(self,
                               P, lam=1, log_trick=True,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_lt_zero.

        Enforce that ``P < 0`` by penalizing invalid solutions with ``lam``.
        See Notes below for more details.

        Parameters
        ----------
        P : dict representing a PUBO.
            The PUBO constraint such that P < 0. Note that ``P`` will be
            converted to a ``qubovert.PUBO`` object if it is not already, thus
            it must follow the conventions, see ``help(qubovert.PUBO)``.
            Please note that if ``P`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        log_trick : bool (optional, defaults to True).
            Whether or not to use the log trick to enforce the inequality
            constraint. See Notes below for more details.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUBO ``P`` can take. If ``bounds`` is None, then they will be
            calculated (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Notes
        -----
        - There is no general way to enforce non integer inequality
          constraints. Thus this function is only guarenteed to work for
          integer inequality constraints (ie constraints of the form
          :math:`x_0 + 2x_1 + ... < 0`). However, it can be used for non
          integer inequality constraints, but it is recommended that the value
          of ``lam`` be set small, since valid solutions may still recieve a
          penalty to the objective function. For example,

          >>> H = PCBO()
          >>> H.add_constraint_lt_zero({(0,): 1, (1,): 2, (2,): -.5, (): .4})
          >>> test_sol = {0: 0, 1: 0, 2: 1}
          >>> H.is_solution_valid(test_sol)
          True
          >>> H.value(test_sol)
          0.01

          {0: 0, 1: 0, 2: 1} is a valid solution to ``H``, but it will still
          cause a nonzero penalty to be added to the objective function.

        - To enforce the inequality constraint, ancilla bits will be
          introduced (labels with `_a`). If ``log_trick`` is ``True``, then
          approximately :math:`\log_2 |\min_x \text{P.value(x)}|`
          ancilla bits will be used. If ``log_trick`` is ``False``, then
          approximately :math:`|\min_x \text{P.value(x)}|` ancilla
          bits will be used.

        Examples
        --------
        Enforce that :math:`-x_a x_b x_c + x_a -4x_a x_b + 3x_c < 2`.

        >>> H = PCBO().add_constraint_lt_zero(
                {('a', 'b', 'c'): -1, ('a',): 1,
                 ('a', 'b'): -4, ('c',): 3, (): -2}
            )
        >>> H
        {('b', 'c', 'a'): -19, ('b', '__a0', 'c', 'a'): -2,
         ('b', 'c', 'a', '__a1'): -4, ('a',): -3, ('b', 'a'): 24,
         ('c', 'a'): 6, ('__a0', 'a'): 2, ('a', '__a1'): 4,
         ('b', '__a0', 'a'): -8, ('b', 'a', '__a1'): -16, ('c',): -3,
         ('__a0', 'c'): 6, ('c', '__a1'): 12, (): 4, ('__a0',): -3,
         ('__a1',): -4, ('__a0', '__a1'): 4}
        >>> H.is_solution_valid({'b': 0, 'c': 0, 'a': 1})
        True
        >>> H.is_solution_valid({'b': 0, 'c': 1, 'a': 1})
        False

        """
        P = PUBO(P)
        self._append_constraint("lt", P)

        min_val, max_val = _get_bounds(P, bounds)

        if min_val >= 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint cannot be satisfied")
            self += lam * P
        elif max_val < 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint is always satisfied")
        else:
            # copy P, don't do +=
            P = P + 1
            min_val += 1
            max_val += 1
            self.add_constraint_le_zero(
                P, lam=lam, log_trick=log_trick,
                bounds=(min_val, max_val), suppress_warnings=True
            )
            self._pop_constraint("le")

        return self

    def add_constraint_le_zero(self,
                               P, lam=1, log_trick=True,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_le_zero.

        Enforce that ``P <= 0`` by penalizing invalid solutions with ``lam``.

        Parameters
        ----------
        P : dict representing a PUBO.
            The PUBO constraint such that P <= 0. Note that ``P`` will be
            converted to a ``qubovert.PUBO`` object if it is not already, thus
            it must follow the conventions, see ``help(qubovert.PUBO)``.
            Please note that if ``P`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        log_trick : bool (optional, defaults to True).
            Whether or not to use the log trick to enforce the inequality
            constraint. See Notes below for more details.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUBO ``P`` can take. If ``bounds`` is None, then they will be
            calculated (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Notes
        -----
        - There is no general way to enforce non integer inequality
          constraints. Thus this function is only guarenteed to work for
          integer inequality constraints (ie constraints of the form
          :math:`x_0 + 2x_1 + ... \leq 0`). However, it can be used for non
          integer inequality constraints, but it is recommended that the value
          of ``lam`` be set small, since valid solutions may still recieve a
          penalty to the objective function. For example,

          >>> H = PCBO()
          >>> H.add_constraint_le_zero({(0,): 1, (1,): 2, (2,): -1.5, (): .4})
          >>> H
          {(0,): 1.7999999999999998, (0, 1): 4, (0, 2): -3.0, (0, '__a0'): 2,
           (1,): 5.6, (1, 2): -6.0, (1, '__a0'): 4, (2,): 1.0499999999999998,
           (2, '__a0'): -3.0, (): 0.16000000000000003, ('__a0',): 1.8}
          >>> test_sol = {0: 0, 1: 0, 2: 1, '__a0': 1}
          >>> H.is_solution_valid(test_sol)
          True
          >>> H.value(test_sol)
          0.01

          {0: 0, 1: 0, 2: 1} is a valid solution to ``H``, but it will still
          cause a nonzero penalty to be added to the objective function.

        - To enforce the inequality constraint, ancilla bits will be
          introduced (labels with `_a`). If ``log_trick`` is ``True``, then
          approximately :math:`\log_2 |\min_x \text{P.value(x)}|`
          ancilla bits will be used. If ``log_trick`` is ``False``, then
          approximately :math:`|\min_x \text{P.value(x)}|` ancilla
          bits will be used.

        Examples
        --------
        Enforce that :math:`-x_a x_b x_c + x_a -4x_a x_b + 3x_c \leq 2`.

        >>> H = PCBO().add_constraint_le_zero(
                {('a', 'b', 'c'): -1, ('a',): 1,
                 ('a', 'b'): -4, ('c',): 3, (): -2}
            )
        >>> H
        {('b', 'c', 'a'): -19, ('b', 'c', 'a', '__a0'): -2,
         ('__a1', 'b', 'c', 'a'): -4, ('__a2', 'b', 'c', 'a'): -8, ('a',): -3,
         ('b', 'a'): 24, ('c', 'a'): 6, ('a', '__a0'): 2, ('__a1', 'a'): 4,
         ('__a2', 'a'): 8, ('b', 'a', '__a0'): -8, ('__a1', 'b', 'a'): -16,
         ('__a2', 'b', 'a'): -32, ('c',): -3, ('c', '__a0'): 6,
         ('__a1', 'c'): 12, ('__a2', 'c'): 24, (): 4, ('__a0',): -3,
         ('__a1',): -4, ('__a1', '__a0'): 4, ('__a2', '__a0'): 8,
         ('__a2', '__a1'): 16}
        >>> H.is_solution_valid({'b': 0, 'c': 0, 'a': 1})
        True
        >>> H.is_solution_valid({'b': 0, 'c': 1, 'a': 1})
        False

        """
        P = PUBO(P)
        self._append_constraint("le", P)

        bounds = min_val, max_val = _get_bounds(P, bounds)
        if _special_constraints_le_zero(self, P, lam, log_trick, bounds):
            return self

        if min_val > 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint cannot be satisfied")
            self += lam * P
        elif max_val <= 0:
            if not suppress_warnings:
                QUBOVertWarning.warn("Constraint is always satisfied")
        else:
            # don't mutate the P that we put in self._constraints
            P = P.copy()
            if min_val:
                for i in range(num_bits(-min_val, log_trick)):
                    v = pow(2, i) if log_trick else 1
                    P[(self._next_ancilla,)] += v
                    max_val += v

            self.add_constraint_eq_zero(
                P, lam=lam,
                bounds=(min_val, max_val), suppress_warnings=True
            )
            self._pop_constraint("eq")

        return self

    def add_constraint_gt_zero(self,
                               P, lam=1, log_trick=True,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_gt_zero.

        Enforce that ``P > 0`` by penalizing invalid solutions with ``lam``.

        Parameters
        ----------
        P : dict representing a PUBO.
            The PUBO constraint such that P > 0. Note that ``P`` will be
            converted to a ``qubovert.PUBO`` object if it is not already, thus
            it must follow the conventions, see ``help(qubovert.PUBO)``.
            Please note that if ``P`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        log_trick : bool (optional, defaults to True).
            Whether or not to use the log trick to enforce the inequality
            constraint. See Notes below for more details.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUBO ``P`` can take. If ``bounds`` is None, then they will be
            calculated (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Notes
        -----
        - There is no general way to enforce non integer inequality
          constraints. Thus this function is only guarenteed to work for
          integer inequality constraints (ie constraints of the form
          :math:`x_0 + 2x_1 + ... > 0`). However, it can be used for non
          integer inequality constraints, but it is recommended that the value
          of ``lam`` be set small, since valid solutions may still recieve a
          penalty to the objective function. For example,

          >>> H = PCBO()
          >>> H.add_constraint_gt_zero({(0,): -1, (1,): -2, (2,): .5, (): -.4})
          >>> test_sol = {0: 0, 1: 0, 2: 1}
          >>> H.is_solution_valid(test_sol)
          True
          >>> H.value(test_sol)
          0.01

          {0: 0, 1: 0, 2: 1} is a valid solution to ``H``, but it will still
          cause a nonzero penalty to be added to the objective function.

        - To enforce the inequality constraint, ancilla bits will be
          introduced (labels with `_a`). If ``log_trick`` is ``True``, then
          approximately :math:`\log_2 |\max_x \text{P.value(x)}|`
          ancilla bits will be used. If ``log_trick`` is ``False``, then
          approximately :math:`|\max_x \text{P.value(x)}|` ancilla
          bits will be used.

        Examples
        --------
        Enforce that :math:`x_a x_b x_c - x_a + 4x_a x_b - 3x_c > -2`.

        >>> H = PCBO().add_constraint_gt_zero(
                {('a', 'b', 'c'): 1, ('a',): -1,
                 ('a', 'b'): 4, ('c',): -3, (): 2}
            )
        >>> H
        {('b', 'c', 'a'): -19, ('b', '__a0', 'c', 'a'): -2,
         ('b', 'c', 'a', '__a1'): -4, ('a',): -3, ('b', 'a'): 24,
         ('c', 'a'): 6, ('__a0', 'a'): 2, ('a', '__a1'): 4,
         ('b', '__a0', 'a'): -8, ('b', 'a', '__a1'): -16, ('c',): -3,
         ('__a0', 'c'): 6, ('c', '__a1'): 12, (): 4, ('__a0',): -3,
         ('__a1',): -4, ('__a0', '__a1'): 4}
        >>> H.is_solution_valid({'b': 0, 'c': 0, 'a': 1})
        True
        >>> H.is_solution_valid({'b': 0, 'c': 1, 'a': 1})
        False

        """
        P = PUBO(P)
        self._append_constraint("gt", P)
        min_val, max_val = _get_bounds(P, bounds)
        bounds = -max_val, -min_val
        self.add_constraint_lt_zero(
            -P, lam=lam, log_trick=log_trick,
            bounds=bounds, suppress_warnings=suppress_warnings
        )
        self._pop_constraint("lt")
        return self

    def add_constraint_ge_zero(self,
                               P, lam=1, log_trick=True,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_ge_zero.

        Enforce that ``P >= 0`` by penalizing invalid solutions with ``lam``.

        Parameters
        ----------
        P : dict representing a PUBO.
            The PUBO constraint such that P >= 0. Note that ``P`` will be
            converted to a ``qubovert.PUBO`` object if it is not already, thus
            it must follow the conventions, see ``help(qubovert.PUBO)``.
            Please note that if ``P`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        log_trick : bool (optional, defaults to True).
            Whether or not to use the log trick to enforce the inequality
            constraint. See Notes below for more details.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUBO ``P`` can take. If ``bounds`` is None, then they will be
            calculated (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Notes
        -----
        - There is no general way to enforce non integer inequality
          constraints. Thus this function is only guarenteed to work for
          integer inequality constraints (ie constraints of the form
          :math:`x_0 + 2x_1 + ... \geq 0`). However, it can be used for non
          integer inequality constraints, but it is recommended that the value
          of ``lam`` be set small, since valid solutions may still recieve a
          penalty to the objective function. For example,

          >>> H = PCBO()
          >>> H.add_constraint_ge_zero({(0,): -1, (1,): -2, (2,):1.5, (): -.4})
          >>> H
          {(0,): 1.7999999999999998, (0, 1): 4, (0, 2): -3.0, (0, '__a0'): 2,
           (1,): 5.6, (1, 2): -6.0, (1, '__a0'): 4, (2,): 1.0499999999999998,
           (2, '__a0'): -3.0, (): 0.16000000000000003, ('__a0',): 1.8}
          >>> test_sol = {0: 0, 1: 0, 2: 1, '__a0': 1}
          >>> H.is_solution_valid(test_sol)
          True
          >>> H.value(test_sol)
          0.01

          {0: 0, 1: 0, 2: 1} is a valid solution to ``H``, but it will still
          cause a nonzero penalty to be added to the objective function.

        - To enforce the inequality constraint, ancilla bits will be
          introduced (labels with `_a`). If ``log_trick`` is ``True``, then
          approximately :math:`\log_2 |\max_x \text{P.value(x)}|`
          ancilla bits will be used. If ``log_trick`` is ``False``, then
          approximately :math:`|\max_x \text{P.value(x)}|` ancilla
          bits will be used.

        Examples
        --------
        Enforce that :math:`x_a x_b x_c - x_a + 4x_a x_b - 3x_c \geq -2`.

        >>> H = PCBO().add_constraint_ge_zero(
                {('a', 'b', 'c'): 1, ('a',): -1,
                 ('a', 'b'): 4, ('c',): -3, (): 2}
            )
        >>> H
        {('b', 'c', 'a'): -19, ('b', 'c', 'a', '__a0'): -2,
         ('__a1', 'b', 'c', 'a'): -4, ('__a2', 'b', 'c', 'a'): -8, ('a',): -3,
         ('b', 'a'): 24, ('c', 'a'): 6, ('a', '__a0'): 2, ('__a1', 'a'): 4,
         ('__a2', 'a'): 8, ('b', 'a', '__a0'): -8, ('__a1', 'b', 'a'): -16,
         ('__a2', 'b', 'a'): -32, ('c',): -3, ('c', '__a0'): 6,
         ('__a1', 'c'): 12, ('__a2', 'c'): 24, (): 4, ('__a0',): -3,
         ('__a1',): -4, ('__a1', '__a0'): 4, ('__a2', '__a0'): 8,
         ('__a2', '__a1'): 16}
        >>> H.is_solution_valid({'b': 0, 'c': 0, 'a': 1})
        True
        >>> H.is_solution_valid({'b': 0, 'c': 1, 'a': 1})
        False

        """
        P = PUBO(P)
        self._append_constraint("ge", P)
        min_val, max_val = _get_bounds(P, bounds)
        bounds = -max_val, -min_val
        self.add_constraint_le_zero(
            -P, lam=lam, log_trick=log_trick,
            bounds=bounds, suppress_warnings=suppress_warnings
        )
        self._pop_constraint("le")
        return self

    def add_constraint_eq_AND(self, a, *variables, lam=1):
        r"""add_constraint_eq_AND.

        Add a penalty to the PCBO that enforces that
        :math:`a = v_0 \land v_1 \land v_2 \land ...`,
        with a penalty factor
        ``lam``, where ``v_1 = variables[0]``, ``v_2 = variables[1]``, ...

        Parameters
        ----------
        a : any hashable object or a dict.
            The label for boolean variable ``a``, or its PUBO representation.
        *variables : arguments.
            Each element of variables is a hashable object or a dict
            (its PUBO representation). They are the label of the boolean
            variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = PCBO()
        >>> H.add_constraint_eq_AND('a', 'b', 'c')  # enforce a == b AND c

        >>> H = PCBO()
        >>> # enforce (a AND b AND c AND d) == 'e'
        >>> H.add_constraint_eq_AND('e', 'a', 'b', 'c', 'd')

        >>> from qubovert import boolean_var, PCBO
        >>> a, b, c = boolean_var('a'), boolean_var('b'), boolean_var('c')
        >>> H = PCBO()
        >>> # enforce that a == b AND c
        >>> H.add_constraint_eq_AND(a, b, c)

        References
        ----------
        https://arxiv.org/pdf/1307.8041.pdf equation 6.

        """
        n = len(variables)
        if n < 2:
            raise ValueError("Must supply at least two variables to AND. "
                             "See ``add_constraint_eq_BUFFER`` for less.")

        a = BUFFER(a)
        b, c = 1, 1
        for v in variables[:n // 2]:
            b *= BUFFER(v)
        for v in variables[n // 2:]:
            c *= BUFFER(v)

        P = 3 * a + b * c - 2 * a * (b + c)

        return self.add_constraint_eq_zero(P, lam, bounds=(0, 3))

    def add_constraint_eq_OR(self, a, *variables, lam=1):
        r"""add_constraint_eq_OR.

        Add a penalty to the PCBO that enforces that
        :math:`v_0 \lor v_1 \lor v_2 \lor ... == a`,
        with a penalty factor ``lam``, where ``v_0 = variables[0]``,
        ``v_1 = variables[1]``, ...

        Parameters
        ----------
        a : any hashable object or a dict.
            The label for boolean variable ``a``, or its PUBO representation.
        *variables : arguments.
            Each element of variables is a hashable object or a dict
            (its PUBO representation). They are the label of the boolean
            variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = PCBO()
        >>> H.add_constraint_eq_OR('a', 'b', 'c')  # enforce a == b OR c

        >>> H = PCBO()
        >>> # enforce a == b OR c OR d
        >>> H.add_constraint_eq_OR('a', 'b', 'c', 'd')

        """
        n = len(variables)
        if n < 2:
            raise ValueError("Must supply at least two variables to OR.")

        a = BUFFER(a)
        if n == 2:
            b, c = BUFFER(variables[0]), BUFFER(variables[1])
            P = a + b + c + b * c - 2 * a * (b + c)
            bounds = 0, 3
        else:
            P = PCBO().add_constraint_NOR(*variables) - a
            bounds = -1, 1

        return self.add_constraint_eq_zero(P, lam=lam, bounds=bounds)

    def add_constraint_eq_XOR(self, a, *variables, lam=1):
        r"""add_constraint_eq_XOR.

        Add a penalty to the PCBO that enforces that
        :math:`v_0 \oplus v_1 \oplus ... == a`
        with a penalty factor ``lam``, where ``v_0 = variables[0]``,
        ``v_1 = variables[1]``, ...

        Parameters
        ----------
        a : any hashable object or a dict.
            The label for boolean variable ``a``, or its PUBO representation.
        *variables : arguments.
            Each element of variables is a hashable object or a dict
            (its PUBO representation). They are the label of the boolean
            variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = PCBO()
        >>> H.add_constraint_eq_XOR('a', 'b', 'c')  # enforce a == b XOR c

        >>> H = PCBO()
        >>> # enforce a == b XOR c XOR d
        >>> H.add_constraint_eq_XOR('a', 'b', 'c', 'd')

        """
        P = PCBO().add_constraint_XNOR(*variables) - BUFFER(a)
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(-1, 1))

    def add_constraint_eq_BUFFER(self, a, b, lam=1):
        r"""add_constraint_eq_BUFFER.

        Add a penalty to the PCBO that enforces that :math:`a == b` with
        a penalty factor ``lam``.

        Parameters
        ----------
        a : any hashable object or a dict.
            The label for boolean variables ``a``, or its PUBO representation.
        b : any hashable object or a dict.
            The label for boolean variables ``b``, or its PUBO representation.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = PCBO()
        >>> H.add_constraint_eq_BUFFER('a', 'b')  # enforce a == b

        >>> from qubovert import PCBO, boolean_var
        >>> a, b = boolean_var('a'), boolean_var('b')
        >>> H = PCBO()
        >>> H.add_constraint_eq_BUFFER(a, b)  # enforce a == b

        """
        P = BUFFER(a) - BUFFER(b)
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(-1, 1))

    def add_constraint_eq_NAND(self, a, *variables, lam=1):
        r"""add_constraint_eq_NAND.

        Add a penalty to the PCBO that enforces that
        :math:`\lnot (v_0 \land v_1 \land v_2 \land ...) == a`, with a
        penalty factor ``lam``, where ``v_0 = variables[0]``,
        ``v_1 = variables[1]``, ...

        Parameters
        ----------
        a : any hashable object or a dict.
            The label for boolean variables ``a``, or its PUBO representation.
        *variables : arguments.
            Each element of variables is a hashable object or a dict
            (its PUBO representation). They are the label of the boolean
            variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = PCBO()
        >>> H.add_constraint_eq_NAND('a', 'b', 'c')  # enforce a == b NAND c

        >>> H = PCBO()
        >>> # enforce a == b NAND c NAND d
        >>> H.add_constraint_eq_NAND('a', 'b', 'c', 'd')

        >>> from qubovert import boolean_var, PCBO
        >>> a, b, c = boolean_var('a'), boolean_var('b'), boolean_var('c')
        >>> # enforce a == b NAND c
        >>> H = PCBO().add_constraint_eq_NAND(a, b, c)

        """
        n = len(variables)
        if n < 2:
            raise ValueError("Must supply at least two variables to NAND. "
                             "See ``add_constraint_eq_NOT`` for less.")
        b, c = 1, 1
        for v in variables[:n // 2]:
            b *= BUFFER(v)
        for v in variables[n // 2:]:
            c *= BUFFER(v)

        P = NOT(a) * (3 - 2 * (b + c)) + b * c
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(0, 3))

    def add_constraint_eq_NOR(self, a, *variables, lam=1):
        r"""add_constraint_eq_NOR.

        Add a penalty to the PCBO that enforces that
        :math:`\lnot(v_0 \lor v_1 \lor v_2 \lor ...) == a`,
        with a penalty factor ``lam``, where ``v_0 = variables[0]``,
        ``v_1 = variables[1]``, ...

        Parameters
        ----------
        a : any hashable object or a dict.
            The label for boolean variable ``a``, or its PUBO representation.
        *variables : arguments.
            Each element of variables is a hashable object or a dict
            (its PUBO representation). They are the label of the boolean
            variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = PCBO()
        >>> H.add_constraint_eq_NOR('a', 'b', 'c')  # enforce a == b NOR c

        >>> H = PCBO()
        >>> # enforce a == b NOR c NOR d
        >>> H.add_constraint_eq_NOR('a', 'b', 'c', 'd')

        >>> from qubovert import boolean_var, PCBO
        >>> # enforce a == b NOR c
        >>> a, b, c = boolean_var('a'), boolean_var('b'), boolean_var('c')
        >>> H.add_constraint_eq_NOR(a, b, c)

        """
        n = len(variables)
        if n < 2:
            raise ValueError("Must supply at least two variables to NOR.")

        a = BUFFER(a)
        if n == 2:
            b, c = BUFFER(variables[0]), BUFFER(variables[1])
            P = 1 - a - b - c + b * c + 2 * a * (b + c)
            bounds = 0, 3
        else:
            P = PCBO().add_constraint_OR(*variables) - a
            bounds = -1, 1

        return self.add_constraint_eq_zero(P, lam=lam, bounds=bounds)

    def add_constraint_eq_XNOR(self, a, *variables, lam=1):
        r"""add_constraint_eq_XNOR.

        Add a penalty to the PCBO that enforces that
        :math:\lnot(`v_0 \oplus v_1 \oplus ...) == a`
        with a penalty factor ``lam``, where ``v_0 = variables[0]``,
        ``v_1 = variables[1]``, ...

        Parameters
        ----------
        a : any hashable object or a dict.
            The label for boolean variable ``a``, or its PUBO representation.
        *variables : arguments.
            Each element of variables is a hashable object or a dict
            (its PUBO representation). They are the label of the boolean
            variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = PCBO()
        >>> H.add_constraint_eq_XNOR('a', 'b', 'c')  # enforce a == b XNOR c

        >>> H = PCBO()
        >>> # enforce a == b XNOR c XNOR d
        >>> H.add_constraint_eq_XNOR('a', 'b', 'c', 'd')

        >>> from qubovert import boolean_var, PCBO
        >>> a, b, c = boolean_var('a'), boolean_var('b'), boolean_var('c')
        >>> # enforce a == b XNOR c
        >>> H = PCBO().add_constraint_eq_XNOR(a, b, c)

        """
        P = PCBO().add_constraint_XOR(*variables) - BUFFER(a)
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(-1, 1))

    def add_constraint_eq_NOT(self, a, b, lam=1):
        r"""NOT.

        Add a penalty to the PCBO that enforces that :math:`\lnot a == b` with
        a penalty factor ``lam``.

        Parameters
        ----------
        a : any hashable object or a dict.
            The label for boolean variable ``a``, or its PUBO representation.
        b : any hashable object or a dict.
            The label for boolean variable ``b``, or its PUBO representation.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = PCBO()
        >>> H.add_constraint_eq_NOT('a', 'b')  # enforce NOT(a) == b

        >>> from qubovert import boolean_var, PCBO
        >>> a, b = boolean_var('a'), boolean_var('b')
        >>> H.add_constraint_eq_NOT(a, b)  # enforce NOT(a) == b

        """
        P = PCBO().add_constraint_BUFFER(a) - BUFFER(b)
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(-1, 1))

    def add_constraint_AND(self, *variables, lam=1):
        r"""add_constraint_AND.

        Add a penalty to the PCBO that is only zero when
        :math:`a \land b \land c \land ...` is True, with a penalty factor
        ``lam``, where ``a = variables[0]``, ``b = variables[1]``, etc.

        Parameters
        ----------
        *variables : arguments.
            Each element of variables is a hashable object or a dict
            (its PUBO representation). They are the label of the boolean
            variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = PCBO()
        >>> H.add_constraint_AND('a', 'b')  # enforce a AND b
        >>> H
        {('b', 'a'): -1, (): 1}

        >>> H = PCBO()
        >>> H.add_constraint_AND('a', 'b', 'c', 'd')
        >>> # enforce a AND b AND c AND d

        >>> from qubovert import boolean_var, PCBO
        >>> a, b = boolean_var('a'), boolean_var('b')
        >>> # enforce a AND b
        >>> H = PCBO().add_constraint_AND(a, b)

        """
        return self.add_constraint_BUFFER(AND(*variables), lam=lam)

    def add_constraint_OR(self, *variables, lam=1):
        r"""add_constraint_OR.

        Add a penalty to the PCBO that is only nonzero when
        :math:`a \lor b \lor c \lor d \lor ...` is True, with a penalty factor
        ``lam``, where ``a = variables[0]``, ``b = variables[1]``, etc.

        Parameters
        ----------
        *variables : arguments.
            Each element of variables is a hashable object or a dict
            (its PUBO representation). They are the label of the boolean
            variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = PCBO()
        >>> H.add_constraint_OR('a', 'b')  # enforce a OR b
        >>> H
        {('a',): -1, ('b',): -1, ('b', 'a'): 1, (): 1}

        >>> H = PCBO()
        >>> H.add_constraint_OR('a', 'b', 'c', 'd')  # enforce a OR b OR c OR d

        >>> from qubovert import boolean_var, PCBO
        >>> a, b = boolean_var('a'), boolean_var('b')
        >>> # enforce a OR b
        >>> H = PCBO().add_constraint_OR(a, b)

        """
        P = 1 - OR(*variables)
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(0, 1))

    def add_constraint_XOR(self, *variables, lam=1):
        r"""add_constraint_XOR.

        Add a penalty to the PCBO that is only nonzero when
        :math:`v_0 \oplus v_1 \oplus ... \oplus v_n` is True, with a penalty
        factor ``lam``, where ``v_0 = variables[0]``, ``v_1 = variables[1]``,
        ..., ``v_n = variables[-1]``. See ``qubovert.sat.XOR`` for the
        XOR convention that qubovert uses for more than two inputs.

        Parameters
        ----------
        *variables : arguments.
            Each element of variables is a hashable object or a dict
            (its PUBO representation). They are the label of the boolean
            variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = PCBO()
        >>> H.add_constraint_XOR('a', 'b')  # enforce a XOR b

        >>> H = PCBO()
        >>> H.add_constraint_XOR('a', 'b', 'c')  # enforce a XOR b XOR c

        >>> from qubovert import boolean_var, PCBO
        >>> a, b = boolean_var('a'), boolean_var('b')
        >>> # enforce a XOR b
        >>> H = PCBO().add_constraint_XOR(a, b)

        """
        P = 1 - XOR(*variables)
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(0, 1))

    def add_constraint_BUFFER(self, a, lam=1):
        r"""add_constraint_BUFFER.

        Add a penalty to the PCBO that is only nonzero when :math:`a == 1` is
        True, with a penalty factor ``lam``.

        Parameters
        ----------
        a : any hashable object or a dict.
            The label for boolean variable ``a``, or its PUBO representation.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = PCBO()
        >>> H.add_constraint_BUFFER('a')  # enforce a

        >>> from qubovert import boolean_var, PCBO
        >>> a = boolean_var('a')
        >>> # enforce a
        >>> H = PCBO().add_constraint_BUFFER(a)

        """
        return self.add_constraint_eq_zero(NOT(a), lam=lam, bounds=(0, 1))

    def add_constraint_NAND(self, *variables, lam=1):
        r"""add_constraint_NAND.

        Add a penalty to the PCBO that is only zero when
        :math:`\lnot (a \land b \land c \land ...)` is True, with a penalty
        factor ``lam``, where ``a = variables[0]``, ``b = variables[1]``, etc.

        Parameters
        ----------
        *variables : arguments.
            Each element of variables is a hashable object or a dict
            (its PUBO representation). They are the label of the boolean
            variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = PCBO()
        >>> H.add_constraint_NAND('a', 'b')  # enforce a NAND b

        >>> H = PCBO()
        >>> H.add_constraint_NAND('a', 'b', 'c', 'd')
        >>> # enforce a NAND b NAND c NAND d

        >>> from qubovert import boolean_var, PCBO
        >>> a, b = boolean_var('a'), boolean_var('b')
        >>> # enforce a NAND b
        >>> H = PCBO().add_constraint_NAND(a, b)

        """
        return self.add_constraint_NOT(AND(*variables), lam=lam)

    def add_constraint_NOR(self, *variables, lam=1):
        r"""add_constraint_NOR.

        Add a penalty to the PCBO that is only nonzero when
        :math:`\lnot(a \lor b \lor c \lor d \lor ...)` is True, with a penalty
        factor ``lam``, where ``a = variables[0]``, ``b = variables[1]``, etc.

        Parameters
        ----------
        *variables : arguments.
            Each element of variables is a hashable object or a dict
            (its PUBO representation). They are the label of the boolean
            variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = PCBO()
        >>> H.add_constraint_NOR('a', 'b')  # enforce a NOR b

        >>> H = PCBO()
        >>> H.add_constraint_NOR('a', 'b', 'c', 'd')
        >>> # enforce a NOR b NOR c NOR d

        >>> from qubovert import boolean_var, PCBO
        >>> a, b = boolean_var('a'), boolean_var('b')
        >>> # enforce a NOR b
        >>> H = PCBO().add_constraint_NOR(a, b)

        """
        P = 1 - PCBO().add_constraint_OR(*variables)
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(0, 1))

    def add_constraint_XNOR(self, *variables, lam=1):
        r"""add_constraint_XNOR.

        Add a penalty to the PCBO that is only nonzero when
        :math:`\lnot(v_0 \oplus v_1 \oplus ... \oplus v_n)` is True, with a
        penalty  factor ``lam``, where ``v_0 = variables[0]``,
        ``v_1 = variables[1]``, ..., ``v_n = variables[-1]``. See
        ``qubovert.sat.XNOR`` for the XNOR convention that qubovert uses for
        more than two inputs.

        Parameters
        ----------
        *variables : arguments.
            Each element of variables is a hashable object or a dict
            (its PUBO representation). They are the label of the boolean
            variables.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = PCBO()
        >>> H.add_constraint_XNOR('a', 'b')  # enforce a XNOR b

        >>> H = PCBO()
        >>> H.add_constraint_XNOR('a', 'b', 'c')  # enforce a XNOR b XNOR c

        >>> from qubovert import boolean_var, PCBO
        >>> a, b = boolean_var('a'), boolean_var('b')
        >>> # enforce a XNOR b
        >>> H = PCBO().add_constraint_XNOR(a, b)

        """
        P = 1 - PCBO().add_constraint_XOR(*variables)
        return self.add_constraint_eq_zero(P, lam=lam, bounds=(0, 1))

    def add_constraint_NOT(self, a, lam=1):
        r"""add_constraint_NOT.

        Add a penalty to the PCBO that is only nonzero when
        :math:`\lnot a` is True, with a penalty factor ``lam``.

        Parameters
        ----------
        a : any hashable object or a dict.
            The label for boolean variables ``a``, or its PUBO representation.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the clause.

        Return
        ------
        self : PCBO.
            Updates the PCBO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        >>> H = PCBO()
        >>> H.add_constraint_NOT('a')  # enforce not a
        >>> H
        {('a',): 1}

        >>> from qubovert import boolean_var, PCBO
        >>> a = boolean_var('a')
        >>> # enforce not a
        >>> H = PCBO().add_constraint_NOT(a)

        """
        return self.add_constraint_eq_zero(BUFFER(a), lam=lam, bounds=(0, 1))
